# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import os
import struct
import sys
import traceback
import time

from MAPI import (
    ECImportContentsChanges, SYNC_E_IGNORE, WrapStoreEntryID,
    SYNC_NORMAL, SYNC_ASSOCIATED, SYNC_CATCHUP, SYNC_UNICODE, IStream,
    STREAM_SEEK_SET, RELOP_GE, RELOP_LT, ECImportHierarchyChanges,
    ECImportAddressbookChanges, MAPI_MAILUSER, SYNC_READ_STATE,
    MSGFLAG_READ,
)
from MAPI.Defs import (
    PpropFindProp
)
from MAPI.Tags import (
    PR_ENTRYID, PR_STORE_ENTRYID, PR_EC_PARENT_HIERARCHYID,
    PR_EC_HIERARCHYID, PR_STORE_RECORD_KEY, PR_CONTENTS_SYNCHRONIZER,
    PR_MESSAGE_DELIVERY_TIME, PR_HIERARCHY_SYNCHRONIZER,
    IID_IECImportAddressbookChanges, IID_IECExportAddressbookChanges,
)
from MAPI.Tags import (
    IID_IExchangeImportContentsChanges, IID_IECImportContentsChanges,
    IID_IExchangeExportChanges, IID_IExchangeImportHierarchyChanges,
    IID_IECImportHierarchyChanges
)
from MAPI.Struct import (
    MAPIError, MAPIErrorNotFound, MAPIErrorNoAccess, SPropValue,
    SPropertyRestriction, SAndRestriction, MAPIErrorNetworkError,
)
from MAPI.Time import (
    unixtime
)

from .compat import benc as _benc, bdec as _bdec

from . import item as _item
try:
    from . import folder as _folder
except ImportError:  # pragma: no cover
    _folder = sys.modules[__package__+'.folder']
try:
    from . import user as _user
except ImportError:  # pragma: no cover
    _user = sys.modules[__package__ + '.user']
try:
    from . import store as _store
except ImportError:  # pragma: no cover
    _store = sys.modules[__package__ + '.store']
try:
    from . import utils as _utils
except ImportError:  # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

TESTING = False
if os.getenv('PYKO_TESTING'):  # env variable used in testset
    TESTING = True


class TrackingHierarchyImporter(ECImportHierarchyChanges):
    def __init__(self, server, importer, stats):
        ECImportHierarchyChanges.__init__(self, [IID_IExchangeImportHierarchyChanges, IID_IECImportHierarchyChanges])

        self.server = server
        self.log = server.log
        self.importer = importer
        self.stats = stats

    def ImportFolderChange(self, props):
        if hasattr(self.importer, 'update'):
            eid = _benc(PpropFindProp(props, PR_ENTRYID).Value)
            folder = _folder.Folder(store=self.importer.store, entryid=eid)
            self.importer.update(folder)

    def ImportFolderDeletion(self, flags, sourcekeys):
        if hasattr(self.importer, 'delete'):
            for sourcekey in sourcekeys:
                folder = _folder.Folder(self.importer.store, _check_mapiobj=False)
                folder._sourcekey = _benc(sourcekey)
                self.importer.delete(folder, flags)

    def UpdateState(self, stream):
        pass


class TrackingContentsImporter(ECImportContentsChanges):
    def __init__(self, server, importer, stats):
        ECImportContentsChanges.__init__(self, [IID_IExchangeImportContentsChanges, IID_IECImportContentsChanges])

        self.server = server
        self.log = server.log
        self.importer = importer
        self.stats = stats
        self.skip = False

    def ImportMessageChangeAsAStream(self, props, flags):
        self.ImportMessageChange(props, flags)

    def ImportMessageChange(self, props, flags):
        if self.skip:
            raise MAPIError(SYNC_E_IGNORE)
        if not hasattr(self.importer, 'update'):
            return
        try:
            entryid = PpropFindProp(props, PR_ENTRYID)
            if self.importer.store:
                mapistore = self.importer.store.mapiobj
            else:
                store_entryid = PpropFindProp(props, PR_STORE_ENTRYID).Value
                store_entryid = WrapStoreEntryID(0, b'zarafa6client.dll', store_entryid[:-4]) + self.server.pseudo_url + b'\x00'
                mapistore = self.server._store2(store_entryid)
            item = _item.Item()
            item.server = self.server
            item.store = _store.Store(mapiobj=mapistore, server=self.server)
            try:
                item.mapiobj = _utils.openentry_raw(
                    mapistore, entryid.Value, 0)
                # TODO properties don't exist?
                props = item.mapiobj.GetProps([PR_EC_HIERARCHYID, PR_EC_PARENT_HIERARCHYID, PR_STORE_RECORD_KEY], 0)
                item.docid = props[0].Value
                item.storeid = _benc(props[2].Value)
                self.importer.update(item, flags)
            # TODO mail already deleted, can we do this in a cleaner way?
            except (MAPIErrorNotFound, MAPIErrorNoAccess):
                self.log.debug('received change for entryid %s, but it could \
not be opened', _benc(entryid.Value))
        except Exception:
            self.log.error('could not process change for entryid %s (%r):', _benc(entryid.Value), props)
            self.log.error(traceback.format_exc())
            if self.stats:
                self.stats['errors'] += 1
        raise MAPIError(SYNC_E_IGNORE)

    def ImportMessageDeletion(self, flags, entries):
        if self.skip or not hasattr(self.importer, 'delete'):
            return
        try:
            for entry in entries:
                item = _item.Item()
                item.server = self.server
                item._sourcekey = _benc(entry)
                self.importer.delete(item, flags)
        except Exception:
            self.log.error('could not process delete for entries: %s', [_benc(entry) for entry in entries])
            self.log.error(traceback.format_exc())
            if self.stats:
                self.stats['errors'] += 1

    def ImportPerUserReadStateChange(self, states):
        if self.skip or not hasattr(self.importer, 'read'):
            return
        try:
            for state in states:
                item = _item.Item()
                item.server = self.server
                if self.importer.store:  # TODO system-wide?
                    item.store = self.importer.store
                item._sourcekey = _benc(state.SourceKey)
                self.importer.read(item, bool(state.ulFlags & MSGFLAG_READ))

        except Exception:
            self.log.error('could not process readstate change')
            self.log.error(traceback.format_exc())
            if self.stats:
                self.stats['errors'] += 1

    def UpdateState(self, stream):
        pass


class TrackingGABImporter(ECImportAddressbookChanges):
    def __init__(self, server, importer):
        ECImportAddressbookChanges.__init__(self, [IID_IECImportAddressbookChanges])
        self.server = server
        self.importer = importer

    def ImportABChange(self, type_, entryid):
        if hasattr(self.importer, 'update') and type_ == MAPI_MAILUSER:
            user = self.server.user(userid=_benc(entryid))
            self.importer.update(user)

    def ImportABDeletion(self, type_, entryid):
        if hasattr(self.importer, 'delete') and type_ == MAPI_MAILUSER:
            user = _user.User(server=self.server)
            user._userid = _benc(entryid)
            self.importer.delete(user)

    def UpdateState(self, stream):
        pass


def state(mapiobj, associated=False):
    exporter = mapiobj.OpenProperty(PR_CONTENTS_SYNCHRONIZER, IID_IExchangeExportChanges, 0, 0)
    if associated:
        exporter.Config(None, SYNC_NORMAL | SYNC_ASSOCIATED | SYNC_CATCHUP, None, None, None, None, 0)
    else:
        exporter.Config(None, SYNC_NORMAL | SYNC_CATCHUP, None, None, None, None, 0)
    steps, step = None, 0
    while steps != step:
        steps, step = exporter.Synchronize(step)
    stream = IStream()
    exporter.UpdateState(stream)
    stream.Seek(0, STREAM_SEEK_SET)
    return _benc(stream.Read(0xFFFFF))


def sync_hierarchy(server, syncobj, importer, state, stats=None):
    importer = TrackingHierarchyImporter(server, importer, stats)
    exporter = syncobj.OpenProperty(PR_HIERARCHY_SYNCHRONIZER, IID_IExchangeExportChanges, 0, 0)

    stream = IStream()
    stream.Write(_bdec(state))
    stream.Seek(0, STREAM_SEEK_SET)

    flags = SYNC_NORMAL | SYNC_UNICODE
    exporter.Config(stream, flags, importer, None, None, None, 0)

    step = 0
    while True:
        (steps, step) = exporter.Synchronize(step)
        if (steps == step):
            break

    exporter.UpdateState(stream)

    stream.Seek(0, STREAM_SEEK_SET)
    return _benc(stream.Read(0xFFFFF))


def sync(server, syncobj, importer, state, max_changes, associated=False, window=None, begin=None, end=None, stats=None):
    log = server.log

    importer = TrackingContentsImporter(server, importer, stats)
    exporter = syncobj.OpenProperty(PR_CONTENTS_SYNCHRONIZER, IID_IExchangeExportChanges, 0, 0)

    stream = IStream()
    stream.Write(_bdec(state))
    stream.Seek(0, STREAM_SEEK_SET)

    restriction = None
    if window:
        # sync window of last N seconds
        propval = SPropValue(PR_MESSAGE_DELIVERY_TIME, unixtime(int(time.time()) - window))
        restriction = SPropertyRestriction(RELOP_GE, PR_MESSAGE_DELIVERY_TIME, propval)

    elif begin or end:
        restrs = []
        if begin:
            propval = SPropValue(PR_MESSAGE_DELIVERY_TIME, unixtime(time.mktime(begin.timetuple())))
            restrs.append(SPropertyRestriction(RELOP_GE, PR_MESSAGE_DELIVERY_TIME, propval))
        if end:
            propval = SPropValue(PR_MESSAGE_DELIVERY_TIME, unixtime(time.mktime(end.timetuple())))
            restrs.append(SPropertyRestriction(RELOP_LT, PR_MESSAGE_DELIVERY_TIME, propval))
        if len(restrs) == 1:
            restriction = restrs[0]
        else:
            restriction = SAndRestriction(restrs)

    flags = SYNC_NORMAL | SYNC_UNICODE | SYNC_READ_STATE
    if associated:
        flags |= SYNC_ASSOCIATED
    try:
        if TESTING and os.getenv('PYKO_TEST_NOT_FOUND'):
            raise MAPIErrorNotFound()
        exporter.Config(stream, flags, importer, restriction, None, None, 0)
    except MAPIErrorNotFound:
        # syncid purged because of 'sync_lifetime' option in server.cfg:
        # get new syncid.
        log.warn("Sync state does not exist on server (anymore); \
requesting new one")

        syncid, changeid = struct.unpack('<II', _bdec(state))
        stream = IStream()
        stream.Write(struct.pack('<II', 0, changeid))
        stream.Seek(0, STREAM_SEEK_SET)

        exporter.Config(stream, flags, importer, restriction, None, None, 0)

    step = retry = changes = 0
    sleep_time = 0.4

    while True:
        try:
            try:
                if TESTING and os.getenv('PYKO_TEST_NETWORK_ERROR') and not importer.skip:
                    raise MAPIErrorNetworkError()
                (steps, step) = exporter.Synchronize(step)
            finally:
                importer.skip = False

            changes += 1
            retry = 0

            if (steps == step) or (max_changes and changes >= max_changes):
                break

        except MAPIError as e:
            log.warn("Received a MAPI error or timeout (error=0x%x, \
retry=%d/5)", e.hr, retry)

            time.sleep(sleep_time)

            if sleep_time < 5.0:
                sleep_time *= 2.0

            if retry < 5:
                retry += 1
            else:
                log.error("Too many retries, skipping change")
                if stats is not None:
                    stats['errors'] += 1

                # in case of a timeout or other issue, try to skip
                # the change after trying several times
                importer.skip = True

                retry = 0

    exporter.UpdateState(stream)

    stream.Seek(0, STREAM_SEEK_SET)

    state = stream.Read(0xFFFFF)

    # because changes may be reordered for efficiency, we are not always
    # linearly following the change journal. so the current state cannot
    # always be represented as a single change id. instead, the current state
    # may contain changes which have been synced, relative to a certain
    # change id (so we have synced until this change id, plus these changes).

    # in pyko though, we always sync until there are no further changes,
    # so this should normally not occur.

    # TODO add an ICS flag to disable reordering!

    if len(state) != 8:
        log.error('sync state %d bytes, expect problems', len(state))

    return _benc(state)


def sync_gab(server, mapistore, importer, state):
    stream = IStream()
    stream.Write(_bdec(state))
    stream.Seek(0, STREAM_SEEK_SET)

    importer = TrackingGABImporter(server, importer)
    exporter = mapistore.OpenProperty(
        PR_CONTENTS_SYNCHRONIZER, IID_IECExportAddressbookChanges, 0, 0)

    exporter.Config(stream, 0, importer)
    steps, step = None, 0
    while steps != step:
        steps, step = exporter.Synchronize(step)

    stream = IStream()
    exporter.UpdateState(stream)

    stream.Seek(0, STREAM_SEEK_SET)
    return _benc(stream.Read(0xFFFFF))



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano.

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI import (
    MAPI_UNICODE, ROW_MODIFY, ROW_ADD, MAPI_MODIFY,
)
from MAPI.Struct import (
    MAPIErrorNotFound, ROWENTRY, SPropValue,
)
from MAPI.Tags import (
    PR_MEMBER_ENTRYID, PR_MEMBER_RIGHTS, PR_MEMBER_ID, PR_ACL_TABLE,
    IID_IExchangeModifyTable,
)

from .compat import (
    bdec as _bdec, benc as _benc
)
from .defs import RIGHT_NAME, NAME_RIGHT
from .errors import NotFoundError
from .log import log_exc

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

def _permissions_dumps(obj, stats=None):
    server = obj.server
    log = server.log

    rows = []
    with log_exc(log, stats):
        acl_table = obj.mapiobj.OpenProperty(PR_ACL_TABLE,
            IID_IExchangeModifyTable, 0, 0)
        table = acl_table.GetTable(0)
        for row in table.QueryRows(2147483647, 0):
            entryid = row[1].Value
            try:
                row[1].Value = (b'user', server.sa.GetUser(entryid,
                    MAPI_UNICODE).Username)
            except MAPIErrorNotFound:
                try:
                    row[1].Value = (b'group', server.sa.GetGroup(entryid,
                        MAPI_UNICODE).Groupname)
                except MAPIErrorNotFound:
                    log.warning("skipping access control entry for unknown \
user/group %s", _benc(entryid))
                    continue
            rows.append(row)
    return _utils.pickle_dumps({
        b'data': rows
    })

def _permissions_loads(obj, data, stats=None):
    server = obj.server
    log = server.log

    with log_exc(log, stats):
        data = _utils.pickle_loads(data)
        if isinstance(data, dict):
            data = data[b'data']
        rows = []
        for row in data:
            try:
                member_type, value = row[1].Value
                if member_type == b'user':
                    entryid = server.user(value).userid
                else:
                    entryid = server.group(value).groupid
                row[1].Value = _bdec(entryid)
                rows.append(row)
            except kopano.NotFoundError:
                log.warning("skipping access control entry for unknown \
user/group '%s'", value)
        acltab = obj.mapiobj.OpenProperty(PR_ACL_TABLE,
            IID_IExchangeModifyTable, 0, MAPI_MODIFY)
        acltab.ModifyTable(0, [ROWENTRY(ROW_ADD, row) for row in rows])

class Permission(object):
    """Permission class

    A permission instance combines a store or folder with a user or group and
    a set of permissions.

    Permissions for a given folder are resolved by following the parent
    chain (and ultimately store), until there is a match on user or group.
    """

    def __init__(self, mapitable, mapirow, server): # TODO fix args
        self.mapitable = mapitable
        self.mapirow = mapirow
        self.server = server

    @property
    def member(self): # TODO companies in addition to users/groups?
        """The associated :class:`User <User>` or :class:`group <Group>`."""
        try:
            return self.server.user(self.server.sa.GetUser(
                self.mapirow[PR_MEMBER_ENTRYID], MAPI_UNICODE).Username)
        except (NotFoundError, MAPIErrorNotFound):
            return self.server.group(self.server.sa.GetGroup(
                self.mapirow[PR_MEMBER_ENTRYID], MAPI_UNICODE).Groupname)

    @property
    def rights(self): # TODO add def roles!!
        """The rights given to the associated member.

        Possible rights:

        *read_items*
        *create_items*
        *create_subfolders*
        *edit_own*,
        *edit_all*
        *delete_own*
        *delete_all*
        *folder_owner*
        *folder_contact*
        *folder_visible*
        """
        r = []
        for right, name in RIGHT_NAME.items():
            if self.mapirow[PR_MEMBER_RIGHTS] & right:
                r.append(name)
        return r

    @rights.setter
    def rights(self, value):
        r = 0
        for name in value:
            try:
                r |= NAME_RIGHT[name]
            except KeyError:
                raise NotFoundError("no such right: '%s'" % name)

        self.mapitable.ModifyTable(0, [ROWENTRY(ROW_MODIFY,
            [SPropValue(PR_MEMBER_ID, self.mapirow[PR_MEMBER_ID]),
             SPropValue(PR_MEMBER_ENTRYID, self.mapirow[PR_MEMBER_ENTRYID]),
             SPropValue(PR_MEMBER_RIGHTS, r)]
        )]) # PR_MEMBER_ID needed, or it becomes ROW_ADD

    def __unicode__(self):
        return "Permission('%s')" % self.member.name

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2018 - 2019 Kopano and its licensors (see LICENSE file)
"""

PidLidEmail1AddressType = 'PT_UNICODE:PSETID_Address:0x8082'
PidLidEmail1DisplayName = 'PT_UNICODE:PSETID_Address:0x8080'
PidLidEmail1EmailAddress = 'PT_UNICODE:PSETID_Address:0x8083'
PidLidEmail1OriginalEntryId = 'PT_BINARY:PSETID_Address:0x8085'
PidLidEmail2AddressType = 'PT_UNICODE:PSETID_Address:0x8092'
PidLidEmail2DisplayName = 'PT_UNICODE:PSETID_Address:0x8090'
PidLidEmail2EmailAddress = 'PT_UNICODE:PSETID_Address:0x8093'
PidLidEmail2OriginalEntryId = 'PT_BINARY:PSETID_Address:0x8095'
PidLidEmail3AddressType = 'PT_UNICODE:PSETID_Address:0x80A2'
PidLidEmail3DisplayName = 'PT_UNICODE:PSETID_Address:0x80A0'
PidLidEmail3EmailAddress = 'PT_UNICODE:PSETID_Address:0x80A3'
PidLidEmail3OriginalEntryId = 'PT_BINARY:PSETID_Address:0x80A5'
PidLidReminderDelta = "PT_LONG:PSETID_Common:0x8501"
PidLidReminderSet = "PT_BOOLEAN:PSETID_Common:0x8503"
PidLidCommonStart = "PT_SYSTIME:PSETID_Common:0x8516"
PidLidCommonEnd = "PT_SYSTIME:PSETID_Common:0x8517"
PidLidUseTnef = "PT_BOOLEAN:PSETID_Common:0x8582"
PidLidAttendeeCriticalChange = "PT_SYSTIME:PSETID_Meeting:0x1"
PidLidWhere = "PT_UNICODE:PSETID_Meeting:0x2"
PidLidGlobalObjectId = "PT_BINARY:PSETID_Meeting:0x3"
PidLidIsSilent = "PT_LONG:PSETID_Meeting:0x4"
PidLidIsRecurring = "PT_BOOLEAN:PSETID_Meeting:0x5"
PidLidIsException = "PT_BOOLEAN:PSETID_Meeting:0xA"
PidLidStartRecurrenceDate = "PT_LONG:PSETID_Meeting:0xD"
PidLidStartRecurrenceTime = "PT_LONG:PSETID_Meeting:0xE"
PidLidEndRecurrenceDate = "PT_LONG:PSETID_Meeting:0xF"
PidLidEndRecurrenceTime = "PT_LONG:PSETID_Meeting:0x10"
PidLidOwnerCriticalChange = "PT_SYSTIME:PSETID_Meeting:0x1A"
PidLidCleanGlobalObjectId = "PT_BINARY:PSETID_Meeting:0x23"
PidLidAppointmentMessageClass = "PT_UNICODE:PSETID_Meeting:0x24"
PidLidMeetingType = "PT_LONG:PSETID_Meeting:0x26"
PidLidAppointmentSequence = "PT_LONG:PSETID_Appointment:0x8201"
PidLidAppointmentSequenceTime = "PT_LONG:PSETID_Appointment:0x8202"
PidLidAppointmentLastSequence = "PT_LONG:PSETID_Appointment:0x8203"
PidLidBusyStatus = "PT_LONG:PSETID_Appointment:0x8205"
PidLidLocation = "PT_UNICODE:PSETID_Appointment:0x8208"
PidLidAppointmentStartWhole = "PT_SYSTIME:PSETID_Appointment:0x820D"
PidLidAppointmentEndWhole = "PT_SYSTIME:PSETID_Appointment:0x820E"
PidLidAppointmentRecur = "PT_BINARY:PSETID_Appointment:0x8216"
PidLidAppointmentStateFlags = "PT_LONG:PSETID_Appointment:0x8217"
PidLidResponseStatus = "PT_LONG:PSETID_Appointment:0x8218"
PidLidSendAsIcal = "PT_BOOLEAN:PSETID_Appointment:0x8200" # XXX undocumented?
PidLidAppointmentReplyTime = "PT_SYSTIME:PSETID_Appointment:0x8220"
PidLidRecurring = "PT_BOOLEAN:PSETID_Appointment:0x8223"
PidLidIntendedBusyStatus = "PT_LONG:PSETID_Appointment:0x8224"
PidLidExceptionReplaceTime = "PT_SYSTIME:PSETID_Appointment:0x8228"
PidLidFInvited = "PT_BOOLEAN:PSETID_Appointment:0x8229"
PidLidAppointmentReplyName = "PT_UNICODE:PSETID_Appointment:0x8230"
PidLidRecurrencePattern = "PT_UNICODE:PSETID_Appointment:0x8232"
PidLidTimeZoneStruct = "PT_BINARY:PSETID_Appointment:0x8233"
PidLidTimeZoneDescription = "PT_UNICODE:PSETID_Appointment:0x8234"
PidLidClipStart = "PT_SYSTIME:PSETID_Appointment:0x8235"
PidLidClipEnd = "PT_SYSTIME:PSETID_Appointment:0x8236"
PidLidToAttendeesString = "PT_UNICODE:PSETID_Appointment:0x823B"
PidLidCcAttendeesString = "PT_UNICODE:PSETID_Appointment:0x823C"
PidLidAppointmentProposedStartWhole = "PT_SYSTIME:PSETID_Appointment:0x8250"
PidLidAppointmentProposedEndWhole = "PT_SYSTIME:PSETID_Appointment:0x8251"
PidLidAppointmentProposedDuration = "PT_LONG:PSETID_Appointment:0x8256"
PidLidAppointmentCounterProposal = "PT_BOOLEAN:PSETID_Appointment:0x8257"
PidLidSideEffects = "PT_LONG:common:0x8510"
PidLidSmartNoAttach = "PT_BOOLEAN:common:0x8514"
PidLidReminderSignalTime = "PT_SYSTIME:common:0x8560"
PidLidAppointmentSubType = "PT_BOOLEAN:PSETID_Appointment:0x8215"
PidLidAppointmentColor = "PT_LONG:PSETID_Appointment:0x8214"
PidLidYomiFirstName = 'PT_UNICODE:PSETID_Address:0x802C'
PidLidYomiLastName = 'PT_UNICODE:PSETID_Address:0x802D'
PidLidYomiCompanyName = 'PT_UNICODE:PSETID_Address:0x802E'
PidLidFileUnder = 'PT_UNICODE:PSETID_Address:0x8005'
PidLidInstantMessagingAddress = 'PT_UNICODE:PSETID_Address:0x8062'
PidLidWorkAddressStreet = 'PT_UNICODE:PSETID_Address:0x8045'
PidLidWorkAddressCity = 'PT_UNICODE:PSETID_Address:0x8046'
PidLidWorkAddressState = 'PT_UNICODE:PSETID_Address:0x8047'
PidLidWorkAddressPostalCode = 'PT_UNICODE:PSETID_Address:0x8048'
PidLidWorkAddressCountry = 'PT_UNICODE:PSETID_Address:0x8049'
PidLidDistributionListMembers = 'PT_MV_BINARY:PSETID_Address:0x8055'
PidLidDistributionListOneOffMembers = 'PT_MV_BINARY:PSETID_Address:0x8054'



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - Kopano and its licensors (see LICENSE file)
"""

import collections
import errno
import grp
import pwd
import logging
import logging.handlers
import multiprocessing
from multiprocessing import Process
import os
import os.path
import signal
import socket
import ssl
import sys
import traceback


try:
    import daemon
except ImportError: # pragma: no cover
    pass

from . import config as _config
from . import log as _log
from . import server as _server
from . import errors as _errors
from . import parser as _parser

def _daemon_helper(func, service, log):
    try:
        if not service or isinstance(service, Service):
            if isinstance(service, Service): # TODO
                service.log_queue = multiprocessing.Queue()
                service.log_queue.cancel_join_thread()
                service.ql = _log.QueueListener(service.log_queue, *service.log.handlers)
                service.ql.start()
            func()
        else:
            func(service)
    finally:
        if log and service:
            log.info('stopping %s', service.name)
        if isinstance(service, Service) and service.ql: # XXX move queue stuff into Service
            service.ql.stop()
            service.log_queue.close()
            service.log_queue.join_thread()

def _daemonize(func, options=None, log=None, config=None,
        service=None):
    uid = gid = None
    if config:
        if config.get('run_as_user'):
            uid = pwd.getpwnam(config.get('run_as_user')).pw_uid
        if config.get('run_as_group'):
            gid = grp.getgrnam(config.get('run_as_group')).gr_gid
    if uid is not None and gid is not None:
        for h in log.handlers:
            if isinstance(h, logging.handlers.WatchedFileHandler):
                os.chown(h.baseFilename, uid, gid)
    with daemon.DaemonContext(
            uid=uid,
            gid=gid,
            working_directory='/',
            files_preserve=[h.stream for h in log.handlers if \
                isinstance(h, logging.handlers.WatchedFileHandler)] \
                    if log else None,
            prevent_core=False,
            detach_process=False,
            stdout=sys.stdout,
            stderr=sys.stderr,
    ):
        _daemon_helper(func, service, log)

class Service:
    """
Encapsulates everything to create a simple Kopano service, such as:

- Locating and parsing a configuration file
- Performing logging, as specifified in the configuration file
- Handling common command-line options (-c, -F)
- Daemonization (if no -F specified)

:param name: name of the service; if for example 'search', the configuration
    file should be called ``/etc/kopano/search.cfg`` or passed with -c
:param config: :class:`Configuration <Config>` to use
:param options: OptionParser instance to get settings from (see :func:`parser`)

"""

    def __init__(self, name, config=None, options=None, args=None,
            logname=None, **kwargs):
        self.name = name
        self.__dict__.update(kwargs)
        if not options:
            options, args = _parser.parser('CSKQUPufmvVFw').parse_args()
        self.options, self.args = options, args
        self.name = name
        self.logname = logname
        self.ql = None
        config2 = _config.CONFIG.copy()
        if config:
            config2.update(config)
        if getattr(options, 'config_file', None):
            # TODO useful during testing.
            # could be generalized with optparse callback?
            options.config_file = os.path.abspath(options.config_file)
        #: Service :class:`configuration <Config>` instance.
        self.config = _config.Config(config2, service=name, options=options)
        self.config.data['server_socket'] = \
            os.getenv("KOPANO_SOCKET") or self.config.data['server_socket']
        if getattr(options, 'worker_processes', None):
            self.config.data['worker_processes'] = options.worker_processes
        # check that this works here or daemon may die silently
        # TODO check run_as_user..?
        #: Service logger.
        self.log = _log.logger(self.logname or self.name,
            options=self.options, config=self.config)
        for msg in self.config.info:
            self.log.info(msg)
        for msg in self.config.warnings:
            self.log.warning(msg)
        if self.config.errors:
            for msg in self.config.errors:
                self.log.error(msg)
            sys.exit(1)
        self.stats = collections.defaultdict(int, {'errors': 0})
        self._server = None

    def main(self):
        raise _errors.Error('Service.main not implemented') # pragma: no cover

    @property
    def server(self):
        """Service :class:`server <Server>` instance."""
        if self._server is None:
            self._server = _server.Server(options=self.options,
                config=self.config.data, log=self.log, service=self,
                _skip_check=True)
        return self._server

    def start(self):
        """Start service."""
        for sig in (signal.SIGTERM, signal.SIGINT):
            signal.signal(sig, lambda *args: sys.exit(-sig))
        # TODO long term, reload config?
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        self.log.info('starting %s', self.logname or self.name)
        try:
            # do not run-as-service (e.g. backup)
            if getattr(self.options, 'service', True):
                _daemonize(self.main, options=self.options, log=self.log,
                    config=self.config, service=self)
            else:
                _daemon_helper(self.main, self, self.log)
        except Exception as e:
            self.log.error(traceback.format_exc())
            sys.exit(1)

class Worker(Process):
    """Service worker class.

    Used by services to spawn multiple worker instances.
    """
    def __init__(self, service, name, **kwargs):
        Process.__init__(self)
        self.daemon = True
        self.name = name
        self.service = service
        self._server = None
        self.__dict__.update(kwargs)
        self.log = logging.getLogger(name=self.name)
        if not self.log.handlers:
            loglevel = _log._loglevel(service.options, service.config)
            formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            qh = _log.QueueHandler(service.log_queue)
            qh.setFormatter(formatter)
            qh.setLevel(loglevel)
            self.log.addHandler(qh)
            self.log.setLevel(loglevel)

    @property
    def server(self):
        """Worker :class:`server <Server>` instance."""
        if self._server is None:
            self._server = _server.Server(options=self.service.options,
                config=self.service.config, log=self.service.log,
                service=self.service, _skip_check=True)
        return self._server

    def main(self):
        raise _errors.Error('Worker.main not implemented') # pragma: no cover

    def run(self):
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
        with _log.log_exc(self.log):
            self.main()

class _ZSocket:
    def __init__(self, addr, ssl_key, ssl_cert):
        self.ssl_key = ssl_key
        self.ssl_cert = ssl_cert
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind(addr)
        self.s.listen(socket.SOMAXCONN)

    def accept(self):
        newsocket, fromaddr = self.s.accept()
        connstream = ssl.wrap_socket(newsocket, server_side=True,
            keyfile=self.ssl_key, certfile=self.ssl_cert)
        return connstream, fromaddr

def server_socket(addr, ssl_key=None, ssl_cert=None, log=None):
    if addr.startswith('file://'):
        addr2 = addr.replace('file://', '')
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        if os.path.exists(addr2):
            os.remove(addr2)
        s.bind(addr2)
        s.listen(socket.SOMAXCONN)
    elif addr.startswith('https://'):
        addr2 = addr.replace('https://', '').split(':')
        addr2 = (addr2[0], int(addr2[1]))
        s = _ZSocket(addr2, ssl_key=ssl_key, ssl_cert=ssl_cert)
    else:
        addr2 = addr.replace('http://', '').split(':')
        addr2 = (addr2[0], int(addr2[1]))
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(addr2)
        s.listen(socket.SOMAXCONN)
    if log:
        log.info('listening on socket %s', addr)
    return s

def client_socket(addr, ssl_cert=None, log=None):
    if addr.startswith('file://'):
        addr2 = addr.replace('file://', '')
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    elif addr.startswith('https://'):
        addr2 = addr.replace('https://', '').split(':')
        addr2 = (addr2[0], int(addr2[1]))
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s = ssl.wrap_socket(s, ca_certs=ssl_cert, cert_reqs=ssl.CERT_REQUIRED)
    else:
        addr2 = addr.replace('http://', '').split(':')
        addr2 = (addr2[0], int(addr2[1]))
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(addr2)
    return s



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano.

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import re
import struct

from MAPI.Defs import DEFINE_OLEGUID, DEFINE_GUID
from MAPI.Struct import MAPINAMEID
from MAPI import Tags
from MAPI.Tags import (
    PS_PUBLIC_STRINGS, respNone, respOrganized, respTentative, respAccepted,
    respDeclined, respNotResponded,
)
from MAPI import MNID_STRING, MAPI_DISTLIST, MNID_ID

try:
    REV_TYPE
except NameError:
    REV_TYPE = {}

    import _MAPICore
    for K, V in _MAPICore.__dict__.items():
        if K.startswith('PT_'):
            REV_TYPE[V] = K

try:
    REV_TAG
except NameError:
    REV_TAG = {}
    for K in sorted(Tags.__dict__, reverse=True):
        if K.startswith('PR_'):
            REV_TAG[Tags.__dict__[K]] = K

PSETID_Appointment = DEFINE_OLEGUID(0x00062002, 0, 0)
PSETID_Task = DEFINE_OLEGUID(0x00062003, 0, 0)
PSETID_Address = DEFINE_OLEGUID(0x00062004, 0, 0)
PSETID_Common = DEFINE_OLEGUID(0x00062008, 0, 0)
PSETID_Log = DEFINE_OLEGUID(0x0006200A, 0, 0)
PSETID_Note = DEFINE_OLEGUID(0x0006200E, 0, 0)
PSETID_Sharing = DEFINE_OLEGUID(0x00062040, 0, 0)
PSETID_PostRss = DEFINE_OLEGUID(0x00062041, 0, 0)
PS_INTERNET_HEADERS = DEFINE_OLEGUID(0x00020386, 0, 0)

PSETID_Archive = DEFINE_GUID(
    0x72e98ebc, 0x57d2, 0x4ab5, 0xb0, 0xaa, 0xd5, 0x0a, 0x7b, 0x53, 0x1c, 0xb9
)
PSETID_Meeting = DEFINE_GUID(
    0x6ED8DA90, 0x450B, 0x101B, 0x98, 0xDA, 0x00, 0xAA, 0x00, 0x3F, 0x13, 0x05
)
PSETID_Kopano_CalDav = DEFINE_GUID(
    0x77536087, 0xcb81, 0x4dc9, 0x99, 0x58, 0xea, 0x4c, 0x51, 0xbe, 0x34, 0x86
)
PSETID_AirSync = DEFINE_GUID(
    0x71035549, 0x0739, 0x4DCB, 0x91, 0x63, 0x00, 0xF0, 0x58, 0x0D, 0xBB, 0xDF
)
PSETID_UnifiedMessaging = DEFINE_GUID(
    0x4442858E, 0xA9E3, 0x4E80, 0xB9, 0x00, 0x31, 0x7A, 0x21, 0x0C, 0xC1, 0x5B
)
PSETID_CONTACT_FOLDER_RECIPIENT = DEFINE_GUID(
    0x0AAA42FE, 0xC718, 0x101A, 0xE8, 0x85, 0x0B, 0x65, 0x1C, 0x24, 0x00, 0x00
)
PSETID_ZMT = DEFINE_GUID(
    0x8acdbf85, 0x4738, 0x4dc4, 0x94, 0xa9, 0xd4, 0x89, 0xa8, 0x3e, 0x5c, 0x41
)
PSETID_CalendarAssistant = DEFINE_GUID(
    0x11000E07, 0xB51B, 0x40D6, 0xAF, 0x21, 0xCA, 0xA8, 0x5E, 0xDA, 0xB1, 0xD0
)
PS_EC_IMAP = DEFINE_GUID(
    0x00f5f108, 0x8e3f, 0x46c7, 0xaf, 0x72, 0x5e, 0x20, 0x1c, 0x23, 0x49, 0xe7
)

PSETID_KC = DEFINE_GUID(
    0x63aed8c8, 0x4049, 0x4b75, 0xbc, 0x88, 0x96, 0xdf, 0x9d, 0x72, 0x3f, 0x2f
)

NAMED_PROPS_INTERNET_HEADERS = [
    MAPINAMEID(PS_INTERNET_HEADERS, MNID_STRING, 'x-original-to'),
]

NAMED_PROPS_ARCHIVER = [
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'store-entryids'),
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'item-entryids'),
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'stubbed'),
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'ref-store-entryid'),
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'ref-item-entryid'),
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'ref-prev-entryid'),
    MAPINAMEID(PSETID_Archive, MNID_STRING, 'flags')
]

NAMED_PROPS_KC = [
    MAPINAMEID(PSETID_KC, MNID_ID, 0x0001),
    MAPINAMEID(PSETID_KC, MNID_ID, 0x0002)
]

NAMED_PROP_CATEGORY = MAPINAMEID(PS_PUBLIC_STRINGS, MNID_STRING, 'Keywords')

GUID_NAMESPACE = {
    PSETID_Archive: 'archive',
    PSETID_Common: 'common',
    PSETID_Appointment: 'appointment',
    PSETID_Task: 'task',
    PSETID_Address: 'address',
    PSETID_Log: 'log',
    PS_INTERNET_HEADERS: 'internet_headers',
    PSETID_Meeting: 'meeting',
    PS_EC_IMAP: 'imap',
    PSETID_Kopano_CalDav: 'kopano_caldav',
    PSETID_AirSync: 'airsync',
    PSETID_Sharing: 'sharing',
    PSETID_PostRss: 'postrss',
    PSETID_UnifiedMessaging: 'unifiedmessaging',
    PSETID_CONTACT_FOLDER_RECIPIENT: 'contact_folder_recipient',
    PSETID_ZMT: 'zmt',
    PSETID_CalendarAssistant: 'calendarassistant',
    PSETID_KC: 'kc',
    PS_PUBLIC_STRINGS: 'public',
}
NAMESPACE_GUID = dict((b, a) for (a, b) in GUID_NAMESPACE.items())

STR_GUID = {
    'PSETID_Common': PSETID_Common,
    'PSETID_Appointment': PSETID_Appointment,
    'PSETID_Meeting': PSETID_Meeting,
    'PSETID_Address': PSETID_Address,
    'PSETID_Archive': PSETID_Archive,
    'PSETID_Task': PSETID_Task,
    'PSETID_Log': PSETID_Log,
    'PS_INTERNET_HEADERS': PS_INTERNET_HEADERS,
    'PS_EC_IMAP': PS_EC_IMAP,
    'PSETID_Kopano_CalDav': PSETID_Kopano_CalDav,
    'PSETID_AirSync': PSETID_AirSync,
    'PSETID_Sharing': PSETID_Sharing,
    'PSETID_PostRss': PSETID_PostRss,
    'PSETID_UnifiedMessaging': PSETID_UnifiedMessaging,
    'PSETID_CONTACT_FOLDER_RECIPIENT': PSETID_CONTACT_FOLDER_RECIPIENT,
    'PSETID_ZMT': PSETID_ZMT,
    'PSETID_CalendarAssistant': PSETID_CalendarAssistant,
    'PSETID_KC': PSETID_KC,
    'PS_PUBLIC_STRINGS': PS_PUBLIC_STRINGS,
}

# TODO copied from common/ECDefs.h
def OBJECTCLASS(__type, __class):
    return (__type << 16) | (__class & 0xFFFF)

OBJECTTYPE_MAILUSER = 1
OBJECTTYPE_CONTAINER = 4
ACTIVE_USER = OBJECTCLASS(OBJECTTYPE_MAILUSER, 1)
NONACTIVE_USER = OBJECTCLASS(OBJECTTYPE_MAILUSER, 2)
CONTAINER_COMPANY = OBJECTCLASS(OBJECTTYPE_CONTAINER, 1)

# TODO copied from msr/main.py
MUIDECSAB = DEFINE_GUID(
    0x50a921ac, 0xd340, 0x48ee, 0xb3, 0x19, 0xfb, 0xa7, 0x53, 0x30, 0x44, 0x25
)
def DEFINE_ABEID(type, id):
    fmt = "4B16s3I4B"
    return struct.pack(fmt, 0, 0, 0, 0, MUIDECSAB, 0, type, id, 0, 0, 0, 0)
EID_EVERYONE = DEFINE_ABEID(MAPI_DISTLIST, 1)

ADDR_PROPS = [
    (Tags.PR_ADDRTYPE_W, Tags.PR_EMAIL_ADDRESS_W, Tags.PR_ENTRYID,
     Tags.PR_DISPLAY_NAME_W, Tags.PR_SEARCH_KEY),
    (Tags.PR_SENDER_ADDRTYPE_W, Tags.PR_SENDER_EMAIL_ADDRESS_W,
     Tags.PR_SENDER_ENTRYID, Tags.PR_SENDER_NAME_W,
     Tags.PR_SENDER_SEARCH_KEY),
    (Tags.PR_RECEIVED_BY_ADDRTYPE_W,
     Tags.PR_RECEIVED_BY_EMAIL_ADDRESS_W, Tags.PR_RECEIVED_BY_ENTRYID,
     Tags.PR_RECEIVED_BY_NAME_W, Tags.PR_RECEIVED_BY_SEARCH_KEY),
    (Tags.PR_ORIGINAL_SENDER_ADDRTYPE_W,
     Tags.PR_ORIGINAL_SENDER_EMAIL_ADDRESS_W,
     Tags.PR_ORIGINAL_SENDER_ENTRYID, Tags.PR_ORIGINAL_SENDER_NAME_W,
     Tags.PR_ORIGINAL_SENDER_SEARCH_KEY),
    (Tags.PR_ORIGINAL_AUTHOR_ADDRTYPE_W,
     Tags.PR_ORIGINAL_AUTHOR_EMAIL_ADDRESS_W,
     Tags.PR_ORIGINAL_AUTHOR_ENTRYID, Tags.PR_ORIGINAL_AUTHOR_NAME_W,
     Tags.PR_ORIGINAL_AUTHOR_SEARCH_KEY),
    (Tags.PR_SENT_REPRESENTING_ADDRTYPE_W,
     Tags.PR_SENT_REPRESENTING_EMAIL_ADDRESS_W,
     Tags.PR_SENT_REPRESENTING_ENTRYID,
     Tags.PR_SENT_REPRESENTING_NAME_W,
     Tags.PR_SENT_REPRESENTING_SEARCH_KEY),
    (Tags.PR_RCVD_REPRESENTING_ADDRTYPE_W,
     Tags.PR_RCVD_REPRESENTING_EMAIL_ADDRESS_W,
     Tags.PR_RCVD_REPRESENTING_ENTRYID,
     Tags.PR_RCVD_REPRESENTING_NAME_W,
     Tags.PR_RCVD_REPRESENTING_SEARCH_KEY),
]

# Common/RecurrenceState.h
# Defines for recurrence exceptions
ARO_SUBJECT = 0x0001
ARO_MEETINGTYPE = 0x0002
ARO_REMINDERDELTA = 0x0004
ARO_REMINDERSET = 0x0008
ARO_LOCATION = 0x0010
ARO_BUSYSTATUS = 0x0020
ARO_ATTACHMENT = 0x0040
ARO_SUBTYPE = 0x0080
ARO_APPTCOLOR = 0x0100
ARO_EXCEPTIONAL_BODY = 0x0200

# location of entryids in PR_IPM_OL2007_ENTRYIDS
RSF_PID_RSS_SUBSCRIPTION = 0x8001
RSF_PID_TODO_SEARCH = 0x8004
RSF_PID_SUGGESTED_CONTACTS = 0x8008

# TODO should we make the names pretty much identical, except for case?
ENGLISH_FOLDER_MAP = {
    'Inbox': 'inbox',
    'Drafts': 'drafts',
    'Outbox': 'outbox',
    'Sent Items': 'sentmail',
    'Deleted Items': 'wastebasket',
    'Junk E-mail': 'junk',
    'Calendar': 'calendar',
    'Contacts': 'contacts',
    'Tasks': 'tasks',
    'Notes': 'notes',
}

UNESCAPED_SLASH_RE = re.compile(r'(?<!\\)/')

RIGHT_NAME = {
    0x1: 'read_items',
    0x2: 'create_items',
    0x80: 'create_subfolders',
    0x8: 'edit_own',
    0x20: 'edit_all',
    0x10: 'delete_own',
    0x40: 'delete_all',
    0x100: 'folder_owner',
    0x200: 'folder_contact',
    0x400: 'folder_visible',
}

NAME_RIGHT = dict((b, a) for (a, b) in RIGHT_NAME.items())

URGENCY = {
    0: 'low',
    1: 'normal',
    2: 'high'
}

REV_URGENCY = dict((b, a) for (a, b) in URGENCY.items())

ASF_MEETING = 1
ASF_RECEIVED = 2
ASF_CANCELED = 4

NR_COLOR = { # PidLidAppointmentColor - only used by old clients
    1: 'red',
    2: 'blue',
    3: 'green',
    4: 'grey',
    5: 'orange',
    6: 'cyan',
    7: 'olive',
    8: 'purple',
    9: 'teal',
    10: 'yellow',
}
COLOR_NR = dict((b, a) for (a, b) in NR_COLOR.items())

FB_STATUS = {
    0: 'free',
    1: 'tentative',
    2: 'busy',
    3: 'outofoffice',
}
STATUS_FB = dict((b, a) for (a, b) in FB_STATUS.items())

RESPONSE_STATUS = {
    respNone: 'None',
    respOrganized: 'Organizer',
    respTentative: 'TentativelyAccepted',
    respAccepted: 'Accepted',
    respDeclined: 'Declined',
    respNotResponded: 'NotResponded',
}

STATUS_RESPONSE = dict((b, a) for (a, b) in RESPONSE_STATUS.items())



# SPDX-License-Identifier: AGPL-3.0-only
import importlib.util
import glob
import os
import sys
import inspect

from plugintemplates import MP_CONTINUE, MP_EXIT

if sys.version_info[0] > 2:
    long = int


class PluginManager(object):
    def __init__(self, plugindir, logger):
        self.logger = logger
        self.plugindir = plugindir
        self.plugins = []

    def loadPlugins(self, baseclass):
        self.logger.logInfo("* Loading plugins started")
        if self.plugindir is None:
            raise Exception('Invalid plugins directory')

        directory = os.path.abspath(self.plugindir)
        if not os.path.isdir(directory):
            self.logger.logInfo(
                "! Plugins directory \"%s\" does not exist."
                "Plugins not loaded." % directory
            )
            return 0

        plugins = glob.glob(directory + "/*.py")

        self.logger.logInfo("** Checking plugins in %s" % directory)
        if directory not in sys.path:
            sys.path.insert(0, directory)

        for p in plugins:
            try:
                name = p.split("/")[-1]
                name = name.split(".py")[0]

                if name.startswith('_'):
                    continue

                self.logger.logDebug("** Inspecting file '%s.py'" % name)

                spec = importlib.util.find_spec(name, directory)

                for k, v in inspect.getmembers(spec.loader.load_module(name)):
                    if not inspect.isclass(v):
                        continue
                    if k.startswith('_'):
                        continue

                    if not issubclass(v, baseclass) or v is baseclass:
                        continue

                    instance = v(self.logger)
                    # fixme, use a plugin meta name ?
                    self.logger.logInfo("** Adding plugin '%s'" % k)
                    self.plugins.append((k, instance))

            except Exception as e:
                self.logger.logError("!-- failed to load: %s" % p)
                self.logger.logError("!-- error: %s " % e)

        self.logger.logInfo("* Loading plugins done")
        return len(self.plugins)

    def processPluginFunction(self, functionname, *args):
        plugins = []
        self.logger.logInfo("* %s processing started" % functionname)

        # get the priority of the plugin function
        for (name, i) in self.plugins:

            if hasattr(i, 'prio' + functionname):
                prio = getattr(i, 'prio' + functionname)
            else:
                prio = 9999

            plugins.append((functionname, prio, name, i))

        # sort the plugins on function name and prio
        sortedlist = sorted(plugins, key=lambda a: (a[0], a[1]))

        retval = MP_CONTINUE,
        for (fname, prio, cname, i) in sortedlist:
            if (fname != functionname):
                continue
            try:
                self.logger.logInfo("** Plugin '%s.%s' priority (%d)" % (
                    cname, fname, prio
                ))

                retval = getattr(i, fname)(*args)

                self.logger.logDebug("** Plugin '%s.%s' return result %s" % (
                    cname, fname, repr(retval)
                ))

                if (type(retval) == int or type(retval) == long):
                    self.logger.logDebug(
                        "!- Plugin '%s.%s' convert return value type "
                        "'%s' to a 'tuple'" % (cname, fname, type(retval))
                    )
                    retval = retval,
                if (type(retval) != tuple):
                    self.logger.logWarn(
                        "!- Plugin '%s.%s' returned a wrong"
                        "return value type '%s' except 'tuple'. "
                        "Fallback on defaults" % (cname, fname, type(retval))
                    )
                    retval = MP_CONTINUE,
                elif (type(retval[0]) != int and type(retval[0]) != long):
                    self.logger.logWarn(
                        "!- Plugin '%s.%s' returned a wrong return value "
                        "type '%s' expect 'int' or 'long' type. Fallback "
                        "on defaults" % (cname, fname, type(retval[0]))
                    )
                    retval = MP_CONTINUE,

                self.logger.logInfo(
                    "** Plugin '%s.%s' returncode %d" % (
                        cname, fname, retval[0]
                    )
                )
            except NotImplementedError as e:
                # Most likely the function is not implemented
                self.logger.logDebug("!---------- %s" % e)
            except Exception as e:
                self.logger.logError("!-- error: %s " % e)
                self.logger.logError(
                    "!- Plugin '%s.%s' call ignored please "
                    "check the plugin" % (cname, fname)
                )
                retval = MP_CONTINUE,

            if (retval[0] == MP_EXIT):
                break

        self.logger.logInfo("* %s processing done" % functionname)
        return retval



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import print_function

from datetime import datetime, timedelta
import sys

import kopano


def get_sender(item):
    '''
    Retrieves the sender from either PR_SENDER_* (item.sender) or when a
    delegate has approved the meeting request on behalf of someone
    the PR_SENT_REPRESENTING_* properties are used.
    '''

    if item.from_.entryid != item.sender.entryid:
        fromname = "'{}' on behalf of '{}'".format(item.sender.name, item.from_.name)
    else:
        fromname = "'{}'".format(item.sender.name)

    return fromname


def process(username, server, entryid):
    user = server.user(username)
    item = user.item(entryid)
    mr = item.meetingrequest

    if mr.is_request:
        server.log.debug("Accepting meeting request tentative from %s", get_sender(item))
        mr.accept(tentative=True, response=False)
    elif mr.is_response:
        server.log.debug("Processing meeting request response from %s", get_sender(item))
        mr.process_response()
    elif mr.is_cancellation:
        server.log.debug("Processing meeting request cancellation from %s", get_sender(item))
        mr.process_cancellation()

    now = datetime.now()

    try:
        server.log.debug("Updating freebusy")
        user.freebusy.publish(now - timedelta(7), now + timedelta(180))
    except kopano.errors.NotFoundError as e:
        server.log.error("Unable to publish freebusy information: %s", str(e))


def main():
    username, config_file, entryid = sys.argv[1:]

    config = kopano.Config(filename=config_file)
    server = kopano.server(config=config)

    process(username, server, entryid)


if __name__ == '__main__':  # pragma: no cover
    main()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2017 - 2019 Kopano and its licensors (see LICENSE file)
"""

import copy
import sys
import weakref

from MAPI import (
    MAPI_MESSAGE, MAPI_FOLDER, MAPI_STORE, MAPIAdviseSink, fnevObjectModified,
    fnevObjectCreated, fnevObjectMoved, fnevObjectCopied, fnevObjectDeleted,
)

from MAPI.Tags import (
    IID_IMAPIAdviseSink,
)

from .compat import benc as _benc, bdec as _bdec
from .errors import NotSupportedError

from . import folder as _folder
from . import item as _item

fnevObjTypeMessage = 0x00010000  # TODO to defs?
fnevObjTypeFolder = 0x00020000
fnevIgnoreCounters = 0x00040000

OBJECT_TYPES = ['folder', 'item']
FOLDER_TYPES = ['mail', 'contacts', 'calendar', 'tasks', 'notes']
EVENT_TYPES = ['created', 'updated', 'deleted']

TRACER = sys.gettrace()

class Notification(object):
    """Notification class

    A notification instance indicates a change to :class:`items <Item>`
    or :class:`folders <Folder>`.
    """

    def __init__(self, mapiobj):
        #: The underlying MAPI notification object
        self.mapiobj = mapiobj
        #: The changed :class:`item <Item>` or :class:`folder <Folder>`.
        self.object = None
        #: The type of the changed object (*item*, *folder*).
        self.object_type = None
        #: The type of change (*created*, *updated*, *deleted*)
        self.event_type = None


def _split(mapiobj, store):
    notif = Notification(mapiobj)

    # object
    if mapiobj.ulObjType == MAPI_MESSAGE:
        item = _item.Item()
        item.store = store
        item.server = store.server
        item._entryid = mapiobj.lpEntryID

        notif.object_type = 'item'
        notif.object = item

    elif mapiobj.ulObjType == MAPI_FOLDER:
        folder = _folder.Folder(store=store, entryid=_benc(mapiobj.lpEntryID),
                                _check_mapiobj=False)

        notif.object = folder
        notif.object_type = 'folder'

    elif mapiobj.ulObjType == MAPI_STORE:
        notif.object = store
        notif.object_type = 'store'

    # event
    if mapiobj.ulEventType in (fnevObjectCreated, fnevObjectCopied):
        notif.event_type = 'created'
        yield notif

    elif mapiobj.ulEventType == fnevObjectModified:
        notif.event_type = 'updated'
        yield notif

    elif mapiobj.ulEventType == fnevObjectDeleted:
        if mapiobj.ulObjType == MAPI_MESSAGE:
            item._folder = store.folder(entryid=_benc(mapiobj.lpParentID))

        notif.event_type = 'deleted'
        yield notif

    # TODO test with folders/add test
    elif mapiobj.ulEventType == fnevObjectMoved:
        notif.event_type = 'created'
        yield notif

        notif = copy.copy(notif)

        if mapiobj.ulObjType == MAPI_MESSAGE:
            item = _item.Item()
            item.store = store
            item.server = store.server
            item._entryid = mapiobj.lpOldID
            notif.object = item
            item._folder = store.folder(entryid=_benc(mapiobj.lpOldParentID))

        elif mapiobj.ulObjType == MAPI_FOLDER:  # TODO mapiobj.lpOldID not set?
            folder = _folder.Folder(store=store,
                                    entryid=_benc(mapiobj.lpEntryID), _check_mapiobj=False)
            notif.object = folder

        notif.event_type = 'deleted'
        yield notif


# TODO can't server filter all this?
def _filter(notifs, folder, event_types, folder_types):
    for notif in notifs:
        if notif.event_type not in event_types:
            continue

        if folder and notif.object_type == 'item' and notif.object.folder != folder:
            continue

        if notif.object_type == 'item' and notif.object.folder and notif.object.folder.type_ not in folder_types:
            continue

        if notif.object_type == 'folder' and notif.object and notif.object.type_ not in folder_types:
            continue

        yield notif

class AdviseSink(MAPIAdviseSink):
    def __init__(self, store, folder, event_types, folder_types, delegate):
        MAPIAdviseSink.__init__(self, [IID_IMAPIAdviseSink])

        self.store = weakref.ref(store) if store is not None else None
        self.folder = weakref.ref(folder) if folder is not None else None
        self.delegate = delegate
        self.event_types = event_types
        self.folder_types = folder_types

    def OnNotify(self, notifications):  # pragma: no cover
        # called from a thread created from C, and as tracing is set for each
        # thread separately from Python, tracing doesn't work here (so coverage
        # also doesn't work).
        #
        # threading.settrace won't help, as the thread is created from C,
        # so we just 'inherit' any used tracer from the 'global' python thread.
        if TRACER:
            sys.settrace(TRACER)
        return self._on_notify(notifications)  # method call to start tracing

    def _on_notify(self, notifications):
        if not hasattr(self.delegate, 'update'):
            return 0
        store = self.store() if self.store else None
        folder = self.folder() if self.folder else None
        for n in notifications:
            for m in _filter(_split(n, store), folder,
                             self.event_types, self.folder_types):
                self.delegate.update(m)
        return 0


def _flags(object_types, event_types):
    flags = fnevObjectModified | fnevObjectCreated | fnevObjectMoved | fnevObjectCopied | fnevObjectDeleted | fnevIgnoreCounters

    if 'folder' in object_types:
        flags |= fnevObjTypeFolder
    if 'item' in object_types:
        flags |= fnevObjTypeMessage

    return flags

def subscribe(store, folder, delegate, object_types=None, folder_types=None,
              event_types=None):

    if not store.server.notifications:
        raise NotSupportedError('server connection does not support \
notifications (try Server(notifications=True))')

    object_types = object_types or OBJECT_TYPES
    folder_types = folder_types or FOLDER_TYPES
    event_types = event_types or EVENT_TYPES

    flags = _flags(object_types, event_types)
    sink = AdviseSink(store, folder, event_types, folder_types, delegate)

    if folder:
        delegate._conn = store.mapiobj.Advise(_bdec(folder.entryid), flags, sink)
    else:
        delegate._conn = store.mapiobj.Advise(None, flags, sink)

def unsubscribe(store, delegate):
    store.mapiobj.Unadvise(delegate._conn)



# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.4.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if getattr(globals().get("__spec__"), "parent", None) or __package__ or "." in __name__:
    from . import _libfreebusy
else:
    import _libfreebusy

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


fbFree = _libfreebusy.fbFree
fbTentative = _libfreebusy.fbTentative
fbBusy = _libfreebusy.fbBusy
fbOutOfOffice = _libfreebusy.fbOutOfOffice
fbKopanoAllBusy = _libfreebusy.fbKopanoAllBusy
class IFreeBusyUpdate(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr

    def PublishFreeBusy(self, arg2):
        return _libfreebusy.IFreeBusyUpdate_PublishFreeBusy(self, arg2)

    def ResetPublishedFreeBusy(self):
        return _libfreebusy.IFreeBusyUpdate_ResetPublishedFreeBusy(self)

    def SaveChanges(self, start, end):
        return _libfreebusy.IFreeBusyUpdate_SaveChanges(self, start, end)
    __swig_destroy__ = _libfreebusy.delete_IFreeBusyUpdate

# Register IFreeBusyUpdate in _libfreebusy:
_libfreebusy.IFreeBusyUpdate_swigregister(IFreeBusyUpdate)
class IEnumFBBlock(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr

    def Next(self, celt):
        return _libfreebusy.IEnumFBBlock_Next(self, celt)

    def Skip(self, celt):
        return _libfreebusy.IEnumFBBlock_Skip(self, celt)

    def Reset(self):
        return _libfreebusy.IEnumFBBlock_Reset(self)

    def Clone(self):
        return _libfreebusy.IEnumFBBlock_Clone(self)

    def Restrict(self, start, end):
        return _libfreebusy.IEnumFBBlock_Restrict(self, start, end)
    __swig_destroy__ = _libfreebusy.delete_IEnumFBBlock

# Register IEnumFBBlock in _libfreebusy:
_libfreebusy.IEnumFBBlock_swigregister(IEnumFBBlock)
class IFreeBusyData(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr

    def EnumBlocks(self, start, end):
        return _libfreebusy.IFreeBusyData_EnumBlocks(self, start, end)

    def GetDelegateInfo(self, arg2):
        return _libfreebusy.IFreeBusyData_GetDelegateInfo(self, arg2)

    def SetFBRange(self, rtmStart, rtmEnd):
        return _libfreebusy.IFreeBusyData_SetFBRange(self, rtmStart, rtmEnd)

    def GetFBPublishRange(self):
        return _libfreebusy.IFreeBusyData_GetFBPublishRange(self)
    __swig_destroy__ = _libfreebusy.delete_IFreeBusyData

# Register IFreeBusyData in _libfreebusy:
_libfreebusy.IFreeBusyData_swigregister(IFreeBusyData)
class IFreeBusySupport(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def Open(self, lpMAPISession, lpMsgStore, bStore):
        return _libfreebusy.IFreeBusySupport_Open(self, lpMAPISession, lpMsgStore, bStore)

    def Close(self):
        return _libfreebusy.IFreeBusySupport_Close(self)

    def LoadFreeBusyData(self, cMax, phrStatus):
        return _libfreebusy.IFreeBusySupport_LoadFreeBusyData(self, cMax, phrStatus)

    def LoadFreeBusyUpdate(self, cUsers, lpData4):
        return _libfreebusy.IFreeBusySupport_LoadFreeBusyUpdate(self, cUsers, lpData4)

    def CommitChanges(self):
        return _libfreebusy.IFreeBusySupport_CommitChanges(self)

    def GetDelegateInfo(self, arg2, arg3):
        return _libfreebusy.IFreeBusySupport_GetDelegateInfo(self, arg2, arg3)

    def SetDelegateInfo(self, arg2):
        return _libfreebusy.IFreeBusySupport_SetDelegateInfo(self, arg2)

    def AdviseFreeBusy(self, arg2):
        return _libfreebusy.IFreeBusySupport_AdviseFreeBusy(self, arg2)

    def GetFBDetailSupport(self, arg2, arg3):
        return _libfreebusy.IFreeBusySupport_GetFBDetailSupport(self, arg2, arg3)

    def HrHandleServerSched(self, arg2):
        return _libfreebusy.IFreeBusySupport_HrHandleServerSched(self, arg2)

    def HrHandleServerSchedAccess(self):
        return _libfreebusy.IFreeBusySupport_HrHandleServerSchedAccess(self)

    def FShowServerSched(self, arg2):
        return _libfreebusy.IFreeBusySupport_FShowServerSched(self, arg2)

    def HrDeleteServerSched(self):
        return _libfreebusy.IFreeBusySupport_HrDeleteServerSched(self)

    def GetFReadOnly(self, arg2):
        return _libfreebusy.IFreeBusySupport_GetFReadOnly(self, arg2)

    def SetLocalFB(self, arg2):
        return _libfreebusy.IFreeBusySupport_SetLocalFB(self, arg2)

    def PrepareForSync(self):
        return _libfreebusy.IFreeBusySupport_PrepareForSync(self)

    def GetFBPublishMonthRange(self, arg2):
        return _libfreebusy.IFreeBusySupport_GetFBPublishMonthRange(self, arg2)

    def PublishRangeChanged(self):
        return _libfreebusy.IFreeBusySupport_PublishRangeChanged(self)

    def CleanTombstone(self):
        return _libfreebusy.IFreeBusySupport_CleanTombstone(self)

    def GetDelegateInfoEx(self, arg2, status, start, end):
        return _libfreebusy.IFreeBusySupport_GetDelegateInfoEx(self, arg2, status, start, end)

    def PushDelegateInfoToWorkspace(self):
        return _libfreebusy.IFreeBusySupport_PushDelegateInfoToWorkspace(self)

    def __init__(self):
        _libfreebusy.IFreeBusySupport_swiginit(self, _libfreebusy.new_IFreeBusySupport())
    __swig_destroy__ = _libfreebusy.delete_IFreeBusySupport

# Register IFreeBusySupport in _libfreebusy:
_libfreebusy.IFreeBusySupport_swigregister(IFreeBusySupport)




# SPDX-License-Identifier: AGPL-3.0-only
from PIL import Image

import io

from MAPI import MAPI_MODIFY, ATTACH_BY_VALUE, ATTACH_EMBEDDED_MSG, MAPI_UNICODE, MAPI_DEFERRED_ERRORS
from MAPI.Util import SPropValue
from MAPI.Tags import (PR_HASATTACH, PR_ATTACH_NUM, PR_ATTACH_LONG_FILENAME_W,
                       PR_ATTACH_METHOD, PR_ATTACH_MIME_TAG_W, PR_ATTACH_DATA_OBJ,
                       PR_DISPLAY_NAME_W, PR_ATTACH_EXTENSION_W, PR_ATTACH_DATA_BIN,
                       PR_ATTACH_FILENAME_W,
                       IID_IStream, IID_IMessage, IID_IAttachment)

from plugintemplates import MP_CONTINUE, IMapiDAgentPlugin


class BMP2PNG(IMapiDAgentPlugin):

    prioPostConverting = 10

    def PostConverting(self, session, addrbook, store, folder, message):
        props = message.GetProps([PR_HASATTACH], 0)

        if not props[0].Value:
            return MP_CONTINUE,

        self.FindBMPAttachmentsAndConvert(message)

        return MP_CONTINUE,

    def FindBMPAttachmentsAndConvert(self, message, embedded=False):
        # Get the attachments
        table = message.GetAttachmentTable(MAPI_UNICODE)
        table.SetColumns([PR_ATTACH_NUM, PR_ATTACH_LONG_FILENAME_W, PR_ATTACH_METHOD, PR_ATTACH_MIME_TAG_W], 0)
        changes = False
        while(True):
            rows = table.QueryRows(10, 0)
            if len(rows) == 0:
                break

            for row in rows:
                if row[0].ulPropTag != PR_ATTACH_NUM or row[2].ulPropTag != PR_ATTACH_METHOD:
                    self.logger.logDebug("!--- Attachment found but the row is incomplete, attachment number or method data is invalid. Data: %s" % (row))
                    continue

                attnum = row[0].Value
                method = row[2].Value
                itype = u''
                if row[3].ulPropTag == PR_ATTACH_MIME_TAG_W:
                    itype = row[3].Value.upper()

                if method == ATTACH_BY_VALUE and itype.find(u'BMP') >= 0:
                    try:
                        self.ConvertBMP2PNG(message, attnum)
                        changes = True
                    except Exception as e:
                        self.logger.logError("!--- [%d] Unable to convert BMP to PNG: %s" % (attnum, e))

                elif (method == ATTACH_EMBEDDED_MSG):
                    self.logger.logDebug("*--- [%d] Embedded message found" % (attnum))

                    attach = message.OpenAttach(attnum, None, MAPI_MODIFY)
                    embeddedmsg = attach.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage, 0, MAPI_MODIFY | MAPI_DEFERRED_ERRORS)

                    if self.FindBMPAttachmentsAndConvert(embeddedmsg, True):
                        changes = True
                        attach.SaveChanges(0)
                else:
                    self.logger.logDebug("*--- [%d] Not supported attachment method %d" % (attnum, method))

        if embedded and changes:
            message.SaveChanges(0)

        return changes

    def ConvertBMP2PNG(self, message, attnum):
        attname = u'Unknown'
        attach = message.OpenAttach(attnum, IID_IAttachment, 0)
        attprops = attach.GetProps([PR_ATTACH_LONG_FILENAME_W, PR_DISPLAY_NAME_W, PR_ATTACH_FILENAME_W], 0)
        stream = attach.OpenProperty(PR_ATTACH_DATA_BIN, IID_IStream, 0, MAPI_MODIFY)

        datain = io.BytesIO(stream.Read(0xFFFFFF))
        dataout = io.BytesIO()

        img = Image.open(datain)
        img.save(dataout, 'PNG')
        stream.SetSize(0)
        stream.Seek(0, 0)

        stream.Write(dataout.getvalue())
        stream.Commit(0)
        props = [SPropValue(PR_ATTACH_MIME_TAG_W, u'image/png'), SPropValue(PR_ATTACH_EXTENSION_W, u'.png')]

        if attprops[0].ulPropTag == PR_ATTACH_LONG_FILENAME_W and attprops[0].Value[-4:].upper() == u'.BMP':
            props.append(SPropValue(PR_ATTACH_LONG_FILENAME_W, attprops[0].Value[0:-3] + u'png'))
            attname = attprops[0].Value

        if attprops[1].ulPropTag == PR_DISPLAY_NAME_W and attprops[1].Value[-4:].upper() == u'.BMP':
            props.append(SPropValue(PR_DISPLAY_NAME_W, attprops[1].Value[0:-3] + u'png'))
            if len(attname) == 0:
                attname = attprops[1].Value

        if attprops[2].ulPropTag == PR_ATTACH_FILENAME_W and attprops[2].Value[-4:].upper() == u'.BMP':
            props.append(SPropValue(PR_ATTACH_FILENAME_W, attprops[2].Value[0:-3] + u'png'))
            if len(attname) == 0:
                attname = attprops[2].Value

        attach.SetProps(props)
        attach.SaveChanges(0)

        self.logger.logDebug("*--- [%d] Attachment '%s' converted to png" % (attnum, attname))

        return 0




# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import warnings

class Error(Exception):
    pass

class ConfigError(Error):
    pass

class DuplicateError(Error):
    pass

class NotFoundError(Error):
    pass

class LogonError(Error):
    pass

class NotSupportedError(Error):
    pass

class ArgumentError(Error):
    pass

class _DeprecationWarning(Warning):
    pass

warnings.simplefilter('default', _DeprecationWarning)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2017 - 2019 Kopano and its licensors (see LICENSE file)
"""

import datetime
import random
import struct
import sys
import time

import libfreebusy

from MAPI import (
    MAPI_UNICODE, MODRECIP_MODIFY, RELOP_EQ, MODRECIP_ADD,
    MAPI_TO, MAPI_BCC, SUPPRESS_RECEIPT, MSGFLAG_READ,
    MSGFLAG_UNSENT, MODRECIP_REMOVE
)

from MAPI.Tags import (
    PR_RECIPIENT_TRACKSTATUS, PR_MESSAGE_RECIPIENTS,
    PR_MESSAGE_CLASS_W, PR_ENTRYID, PR_SENT_REPRESENTING_ENTRYID, PR_ROWID,
    PR_ADDRTYPE_W, PR_EMAIL_ADDRESS_W, PR_SMTP_ADDRESS_W, PR_PROCESSED,
    IID_IMAPITable, IID_IMessage, PR_RECIPIENT_TRACKSTATUS_TIME,
    PR_RCVD_REPRESENTING_NAME_W, PR_DISPLAY_NAME_W, PR_RECIPIENT_ENTRYID,
    PR_RECIPIENT_TYPE, PR_SEND_INTERNET_ENCODING, PR_SEND_RICH_INFO,
    PR_RECIPIENT_DISPLAY_NAME, PR_DISPLAY_TYPE, PR_RECIPIENT_FLAGS,
    PR_OBJECT_TYPE, PR_SEARCH_KEY, PR_PROPOSED_NEWTIME, PR_OWNER_APPT_ID,
    PR_PROPOSED_NEWTIME_START, PR_PROPOSED_NEWTIME_END,
    PR_SENT_REPRESENTING_NAME_W, PR_SENT_REPRESENTING_EMAIL_ADDRESS_W,
    PR_RECIPIENT_DISPLAY_NAME_W, PR_SENT_REPRESENTING_ADDRTYPE_W,
    PR_SENT_REPRESENTING_SEARCH_KEY, PR_ACCOUNT_W, PR_DISPLAY_TYPE_EX,
    PR_SUBJECT_W, PR_MESSAGE_FLAGS, PR_RESPONSE_REQUESTED,
    recipSendable, recipOrganizer, recipOriginal, respTentative, respAccepted,
    respDeclined, respNotResponded, PR_START_DATE, PR_END_DATE,
)

from MAPI.Defs import (
    PpropFindProp,
)

from MAPI.Struct import (
    SPropValue, SPropertyRestriction, MAPIErrorUnknownEntryid,
)

from MAPI.Time import NANOSECS_BETWEEN_EPOCH

from .defs import (
    ASF_MEETING, ASF_RECEIVED, ASF_CANCELED
)

from .errors import Error
from .restriction import Restriction

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

from . import property_ as _prop

# TODO move all pidlids into separate definition file, plus short description
# of their meanings

from .pidlid import (
    PidLidReminderDelta, PidLidReminderSet, PidLidCommonStart,
    PidLidCommonEnd, PidLidUseTnef, PidLidAttendeeCriticalChange, PidLidWhere,
    PidLidGlobalObjectId, PidLidIsSilent, PidLidIsRecurring,
    PidLidIsException, PidLidStartRecurrenceDate, PidLidStartRecurrenceTime,
    PidLidEndRecurrenceDate, PidLidEndRecurrenceTime,
    PidLidOwnerCriticalChange, PidLidCleanGlobalObjectId,
    PidLidAppointmentMessageClass, PidLidMeetingType,
    PidLidAppointmentSequence, PidLidAppointmentSequenceTime,
    PidLidAppointmentLastSequence, PidLidBusyStatus, PidLidLocation,
    PidLidAppointmentStartWhole, PidLidAppointmentEndWhole,
    PidLidAppointmentRecur, PidLidAppointmentStateFlags, PidLidResponseStatus,
    PidLidAppointmentReplyTime, PidLidRecurring, PidLidIntendedBusyStatus,
    PidLidExceptionReplaceTime, PidLidFInvited, PidLidAppointmentReplyName,
    PidLidRecurrencePattern, PidLidTimeZoneStruct, PidLidTimeZoneDescription,
    PidLidToAttendeesString, PidLidCcAttendeesString,
    PidLidAppointmentProposedStartWhole, PidLidAppointmentProposedEndWhole,
    PidLidAppointmentProposedDuration, PidLidAppointmentCounterProposal,
    PidLidSendAsIcal,
)

# all of the above # TODO redundant
PROPTAGS = [
    PidLidReminderDelta, PidLidReminderSet, PidLidCommonStart,
    PidLidCommonEnd, PidLidUseTnef, PidLidAttendeeCriticalChange, PidLidWhere,
    PidLidGlobalObjectId, PidLidIsSilent, PidLidIsRecurring,
    PidLidIsException, PidLidStartRecurrenceDate, PidLidStartRecurrenceTime,
    PidLidEndRecurrenceDate, PidLidEndRecurrenceTime,
    PidLidOwnerCriticalChange, PidLidCleanGlobalObjectId,
    PidLidAppointmentMessageClass, PidLidMeetingType,
    PidLidAppointmentSequence, PidLidAppointmentSequenceTime,
    PidLidAppointmentLastSequence, PidLidBusyStatus, PidLidLocation,
    PidLidAppointmentStartWhole, PidLidAppointmentEndWhole,
    PidLidAppointmentRecur, PidLidAppointmentStateFlags, PidLidResponseStatus,
    PidLidAppointmentReplyTime, PidLidRecurring, PidLidIntendedBusyStatus,
    PidLidExceptionReplaceTime, PidLidFInvited, PidLidAppointmentReplyName,
    PidLidRecurrencePattern, PidLidTimeZoneStruct, PidLidTimeZoneDescription,
    PidLidToAttendeesString, PidLidCcAttendeesString,
    PidLidAppointmentProposedStartWhole, PidLidAppointmentProposedEndWhole,
    PidLidAppointmentProposedDuration, PidLidAppointmentCounterProposal,
    PidLidSendAsIcal,
]

RECIP_PROPS = [
    PR_ENTRYID,
    PR_DISPLAY_NAME_W,
    PR_EMAIL_ADDRESS_W,
    PR_RECIPIENT_ENTRYID,
    PR_RECIPIENT_TYPE,
    PR_SEND_INTERNET_ENCODING,
    PR_SEND_RICH_INFO,
    PR_RECIPIENT_DISPLAY_NAME, # TODO _W?
    PR_ADDRTYPE_W,
    PR_DISPLAY_TYPE,
    PR_RECIPIENT_TRACKSTATUS,
    PR_RECIPIENT_TRACKSTATUS_TIME,
    PR_RECIPIENT_FLAGS,
    PR_ROWID,
    PR_OBJECT_TYPE,
    PR_SEARCH_KEY,
]


def _organizer_props(cal_item, item):
    has_organizer = False

    table = cal_item.mapiobj.OpenProperty(
        PR_MESSAGE_RECIPIENTS, IID_IMAPITable, MAPI_UNICODE, 0)
    for row in table.QueryRows(2147483647, 0):
        recipient_flags = PpropFindProp(row, PR_RECIPIENT_FLAGS)
        if recipient_flags and recipient_flags.Value == (recipOrganizer | recipSendable):
            has_organizer = True
            break

    if not has_organizer:
        orgprops = [
            SPropValue(PR_ENTRYID,
                item.prop(PR_SENT_REPRESENTING_ENTRYID).value),
            SPropValue(PR_DISPLAY_NAME_W,
                       item.prop(PR_SENT_REPRESENTING_NAME_W).value),
            SPropValue(PR_EMAIL_ADDRESS_W,
                       item.prop(PR_SENT_REPRESENTING_EMAIL_ADDRESS_W).value),
            SPropValue(PR_RECIPIENT_TYPE, MAPI_TO),
            SPropValue(PR_RECIPIENT_DISPLAY_NAME_W,
                       item.prop(PR_SENT_REPRESENTING_NAME_W).value),
            SPropValue(PR_ADDRTYPE_W,
                       item.prop(PR_SENT_REPRESENTING_ADDRTYPE_W).value),  # TODO php
            SPropValue(PR_RECIPIENT_TRACKSTATUS, 0),
            SPropValue(PR_RECIPIENT_FLAGS, (recipOrganizer | recipSendable)),
        ]
        # TODO not in exception message?
        repr_search_key = item.get(PR_SENT_REPRESENTING_SEARCH_KEY)
        if repr_search_key:
            orgprops.append(SPropValue(PR_SEARCH_KEY, repr_search_key))
        return orgprops


def _copytags(mapiobj):
    copytags = [_prop._name_to_proptag(tag, mapiobj)[0] for tag in PROPTAGS]
    copytags.extend([
        PR_SUBJECT_W,
        PR_SENT_REPRESENTING_NAME_W,
        PR_SENT_REPRESENTING_ADDRTYPE_W,
        PR_SENT_REPRESENTING_EMAIL_ADDRESS_W,
        PR_SENT_REPRESENTING_ENTRYID,
        PR_SENT_REPRESENTING_SEARCH_KEY,
        PR_RCVD_REPRESENTING_NAME_W,
    ])
    return copytags


# TODO: in MAPI.Time?
def mapi_time(t):
    return int(t) * 10000000 + NANOSECS_BETWEEN_EPOCH


def _generate_goid():
    """
    Generate a meeting request Global Object ID.

    The Global Object ID is a MAPI property that any MAPI client uses to
    correlate meeting updates and responses with a particular meeting on
    the calendar. The Global Object ID is the same across all copies of the
    item.

    The Global Object ID consists of the following data:

    * byte array id (16 bytes) - identifiers the BLOB as a GLOBAL Object ID.
    * year (YH + YL) - original year of the instance represented by the
    exception. The value is in big-endian format.
    * M (byte) - original month of the instance represented by the exception.
    * D (byte) - original day of the instance represented by the exception.
    * Creation time - 8 byte date
    * X - reversed byte array of size 8.
    * size - LONG, the length of the data field.
    * data - a byte array (16 bytes) that uniquely identifers the meeting
    object.
    """

    # byte array id
    goid = b'\x04\x00\x00\x00\x82\x00\xe0\x00t\xc5\xb7\x10\x1a\x82\xe0\x08'
    # YEARHIGH, YEARLOW, MONTH, DATE
    goid += struct.pack('>H2B', 0, 0, 0)
    # Creation time, lowdatetime, highdatetime
    now = mapi_time(time.time())
    goid += struct.pack('II', now & 0xffffffff, now >> 32)
    # Reserved, 8 zeros
    goid += struct.pack('L', 0)
    # data size
    goid += struct.pack('I', 16)
    # Unique data
    for _ in range(0, 16):
        goid += struct.pack('B', random.getrandbits(8))
    return goid


def _create_meetingrequest(cal_item, item, basedate=None):
    # TODO Update the calendar item, for tracking status
    # TODO Set the body of the message like WebApp / OL does.
    # TODO Whitelist properties?

    item2 = item.copy(item.store.outbox)
    cancel = item.canceled

    # remove meeting organizer TODO just copy correct ones? or why is the
    # organizer MAPI_TO?
    table = item2.mapiobj.OpenProperty(
        PR_MESSAGE_RECIPIENTS, IID_IMAPITable, MAPI_UNICODE, 0)
    table.SetColumns(RECIP_PROPS, 0)
    orgs = []
    for row in table.QueryRows(2147483647, 0):
        recipient_flags = PpropFindProp(row, PR_RECIPIENT_FLAGS)
        if recipient_flags and recipient_flags.Value & recipOrganizer:
            orgs.append(row)
    item2.mapiobj.ModifyRecipients(MODRECIP_REMOVE, orgs)

    # set meeting request props
    if cancel:
        item2.message_class = 'IPM.Schedule.Meeting.Canceled'
    else:
        item2.message_class = 'IPM.Schedule.Meeting.Request'

    stateflags = ASF_MEETING | ASF_RECEIVED
    if cancel:
        stateflags |= ASF_CANCELED
        item2.subject = 'Canceled: '+item2.subject
    item2[PidLidAppointmentStateFlags] = stateflags

    # create appointment goid if not there
    cleangoid = cal_item.get(PidLidCleanGlobalObjectId)
    if cleangoid is None:
        cleangoid = _generate_goid()
        cal_item[PidLidGlobalObjectId] = cleangoid
        cal_item[PidLidCleanGlobalObjectId] = cleangoid

    # update sequence props
    sequence = cal_item.get(PidLidAppointmentSequence)
    if sequence is None:
        cal_item[PidLidAppointmentSequence] = 0
        cal_item[PidLidAppointmentLastSequence] = 0
    else:
        cal_item[PidLidAppointmentSequence] = sequence+1
        cal_item[PidLidAppointmentLastSequence] = sequence+1
    item2[PidLidAppointmentSequence] = \
        cal_item[PidLidAppointmentSequence]
    item2[PidLidAppointmentLastSequence] = \
        cal_item[PidLidAppointmentLastSequence]
    item2[PidLidResponseStatus] = respNotResponded

    # set item goids
    item2[PidLidCleanGlobalObjectId] = cleangoid
    if basedate:
        datefield = struct.pack('>H2B',
            basedate.year, basedate.month, basedate.day)
        goid = cleangoid[:16] + datefield + cleangoid[20:]

        item2[PidLidGlobalObjectId] = goid
        item2[PidLidRecurring] = False

        # update for non-exception TODO don't overwrite if exception
        if cancel:
            item2[PidLidAppointmentStartWhole] = basedate
            item2[PidLidAppointmentEndWhole] = \
                basedate + (cal_item.end - cal_item.start)

            item2[PR_START_DATE] = basedate
            item2[PR_END_DATE] = basedate + (cal_item.end - cal_item.start)

            item2[PidLidExceptionReplaceTime] = \
                datetime.datetime(basedate.year, basedate.month, basedate.day)

    return item2


class MeetingRequest(object):
    """MeetingRequest class

    A meeting request is basically a copy of an appointment, being
    sent to the attendees (or sent back to the organizer).

    It may be an actual request, response or cancellation for the
    given appointment.
    """

    def __init__(self, item):
        self.item = item
        self.log = item.server.log

    @property
    def calendar(self):
        """Respective :class:`calendar <Folder>`
        (possibly in delegator store)."""
        delegator = self.item.get_prop(PR_RCVD_REPRESENTING_NAME_W)
        if delegator:
            store = self.item.server.user(delegator.value).store
        else:
            store = self.item.store # TODO pass to __init__

        return store.calendar

    # TODO ambiguous: split in two (match exact GOID or parent recurrence?)
    @property
    def calendar_item(self):
        """Global calendar item :class:`item <Item>`
        (possibly in delegator store)."""
        goid = self.item.get_prop(PidLidCleanGlobalObjectId)
        if goid is not None:
            restriction = Restriction(SPropertyRestriction(
                RELOP_EQ, goid.proptag,
                SPropValue(goid.proptag, goid.mapiobj.Value))
            )
            return next(self.calendar.items(restriction), None)

    @property
    def basedate(self):
        """Exception date."""
        blob = self.item.get(PidLidGlobalObjectId)
        if blob is not None:
            y, m, d = struct.unpack_from('>HBB', blob, 16)
            if (y, m, d) != (0, 0, 0):
                return datetime.datetime(y, m, d)

    @property
    def update_counter(self):
        """Update counter."""
        return self.item.get(PidLidAppointmentSequence)

    @property
    def track_status(self):  # TODO use strs
        return {
            'IPM.Schedule.Meeting.Resp.Tent': respTentative,
            'IPM.Schedule.Meeting.Resp.Pos': respAccepted,
            'IPM.Schedule.Meeting.Resp.Neg': respDeclined,
        }.get(self.item.message_class)

    @property
    def processed(self):
        """Has the request/response been processed."""
        return self.item.get(PR_PROCESSED, False)

    @processed.setter
    def processed(self, value):
        self.item[PR_PROCESSED] = value

    def _check_processed(self):
        if self.processed:
            return True
        else:
            self.processed = True

    @property
    def response_requested(self):
        """Is a response requested."""
        return self.item.get(PR_RESPONSE_REQUESTED, False)

    @property
    def is_request(self):
        """Is it a request."""
        return self.item.message_class == 'IPM.Schedule.Meeting.Request'

    @property
    def is_response(self):
        """Is it a response."""
        return self.item.message_class.startswith(
            'IPM.Schedule.Meeting.Resp.')

    @property
    def is_cancellation(self):
        """Is it a cancellation."""
        return self.item.message_class == 'IPM.Schedule.Meeting.Canceled'

    def _respond(self, subject_prefix, message_class, message=None):
        response = self.item.copy(self.item.store.outbox)
        response.message_class = message_class

        response.subject = subject_prefix + ': ' + self.item.subject
        if message:
            response.text = message
        # TODO: error handling when user is not found
        response.to = self.item.server.user(email=self.item.from_.email)
        response.from_ = self.item.store.user  # TODO slow?

        response.send()

    # TODO what about updating existing
    def _init_calitem(self, cal_item, tentative, merge=False):
        cal_item.message_class = 'IPM.Appointment'

        # busystatus
        intended_busystatus = self.item.get(PidLidIntendedBusyStatus)
        if intended_busystatus is not None:
            if tentative and intended_busystatus != libfreebusy.fbFree:
                busystatus = libfreebusy.fbTentative
            else:
                busystatus = intended_busystatus
            cal_item[PidLidBusyStatus] = busystatus

        # add organizer as recipient
        organizer_props = _organizer_props(cal_item, self.item)
        if organizer_props and not merge:
            table = cal_item.mapiobj.OpenProperty(
                PR_MESSAGE_RECIPIENTS, IID_IMAPITable, MAPI_UNICODE, 0)
            table.SetColumns(RECIP_PROPS, 0)
            rows = table.QueryRows(2147483647, 0)
            # TODO php-compat: php checks 'move' flag, should we
            if tentative:
                cal_item.mapiobj.ModifyRecipients(MODRECIP_REMOVE, rows)
                cal_item.mapiobj.ModifyRecipients(MODRECIP_ADD, [organizer_props] + rows)
            else:
                cal_item.mapiobj.ModifyRecipients(MODRECIP_ADD, [organizer_props])
            _utils._save(cal_item.mapiobj)

    def accept(self, tentative=False, response=True, add_bcc=False, subject_prefix='Accepted'):
        """Accept meeting request.

        :param tentative: accept tentatively (default False)
        :param response: send response message (default True)
        """
        if not self.is_request:
            raise Error('item is not a meeting request')
        if self._check_processed():
            self.log.warning('meeting request already processed')
            return

        calendar = self.calendar
        cal_item = self.calendar_item
        basedate = self.basedate

        # do nothing for invitation to self
        # (esp. don't try to add organizer/BCC)
        if self.item.store.user.email == self.item.from_.email:
            return

        self.item.mapiobj.SetReadFlag(SUPPRESS_RECEIPT)
        merge = False

        if basedate:
            # update existing recurrence
            if cal_item and cal_item.recurring:
                recurrence = cal_item.recurrence
                if recurrence._is_exception(basedate):
                    recurrence._modify_exception(basedate, self.item)
                else:
                    recurrence._create_exception(basedate, self.item)

            # otherwise replace calendar item
            else:
                if cal_item and cal_item[PidLidGlobalObjectId] == \
                        self.item[PidLidGlobalObjectId]:
                    calendar.delete(cal_item)
                cal_item = self.item.copy(calendar)
                self._init_calitem(cal_item, tentative)

        else:
            # preserve categories # TODO preserve everything unrelated?
            categories = cal_item.categories if cal_item else []

            # remove existing recurrence
            if cal_item and cal_item.recurring:
                mr_counter = self.update_counter
                cal_counter = cal_item.meetingrequest.update_counter
                if mr_counter is not None and cal_counter is not None and \
                   mr_counter <= cal_counter:
                    raise Error('trying to accept out-of-date meeting request')
                calendar.delete(cal_item)

            # determine existing exceptions
            goid = self.item.get_prop(PidLidCleanGlobalObjectId)
            if goid is not None:
                restriction = Restriction(SPropertyRestriction(
                    RELOP_EQ, goid.proptag,
                    SPropValue(goid.proptag, goid.mapiobj.Value))
                )
                existing_items = list(calendar.items(restriction))
                # TODO check php
                existing_items.sort(
                    key=lambda i: i.get(PidLidAppointmentStartWhole))
            else:
                existing_items = []

            # create new recurrence
            cal_item = self.item.copy(calendar)
            if categories:
                cal_item.categories = categories

            # merge existing exceptions
            if cal_item.recurring and existing_items:
                merge = True
                rec = cal_item.recurrence
                for item in existing_items:
                    if item.meetingrequest.basedate and not rec._is_exception(item.meetingrequest.basedate):
                        rec._create_exception(
                            item.meetingrequest.basedate, item, merge=True)
                    # TODO else update..?

            calendar.delete(existing_items)

            self._init_calitem(cal_item, tentative, merge)

            # TODO why filter recipients?
            if merge or not add_bcc:
                table = cal_item.mapiobj.OpenProperty(
                    PR_MESSAGE_RECIPIENTS, IID_IMAPITable, MAPI_UNICODE, 0)
                table.SetColumns(RECIP_PROPS, 0)
                rows = table.QueryRows(2147483647, 0)
                cal_item.mapiobj.ModifyRecipients(MODRECIP_MODIFY, rows)
                _utils._save(cal_item.mapiobj)

        # add self as BCC for ZCP-9901 TODO still relevant?
        proptags = [
            PR_ACCOUNT_W,
            PR_ADDRTYPE_W,
            PR_DISPLAY_NAME_W,
            PR_DISPLAY_TYPE,
            PR_DISPLAY_TYPE_EX,
            PR_EMAIL_ADDRESS_W,
            PR_ENTRYID,
            PR_OBJECT_TYPE,
            PR_SEARCH_KEY,
            PR_SMTP_ADDRESS_W,
        ]
        user = self.item.store.user
        props = user.mapiobj.GetProps(proptags, 0)
        props.extend([
            SPropValue(PR_RECIPIENT_ENTRYID, props[6].Value),
            SPropValue(PR_RECIPIENT_FLAGS, (recipOriginal | recipSendable)),
            SPropValue(PR_RECIPIENT_TRACKSTATUS, 0),
            SPropValue(PR_RECIPIENT_TYPE, MAPI_BCC),
        ])
        if add_bcc and not merge:
            cal_item.mapiobj.ModifyRecipients(MODRECIP_ADD, [props])
            _utils._save(cal_item.mapiobj)

        # send response
        if response:
            if tentative:
                message_class = 'IPM.Schedule.Meeting.Resp.Tent'
            else:
                message_class = 'IPM.Schedule.Meeting.Resp.Pos'
            self._respond(subject_prefix, message_class)

    def decline(self, message=None, response=True, subject_prefix='Declined'):
        """Decline meeting request.

        :param response: send response message (default True)
        """
        message_class = 'IPM.Schedule.Meeting.Resp.Neg'

        if response:
            self._respond(subject_prefix, message_class, message)

    def process_cancellation(self, delete=False):
        """Process meeting request cancellation.

        :param delete: delete appointment from calendar (default False)
        """

        if not self.is_cancellation:
            raise Error('item is not a meeting request cancellation')

        cal_item = self.calendar_item
        if not cal_item:
            self.log.debug('no appointment matches cancellation')
            return

        basedate = self.basedate

        if basedate:
            if cal_item.recurring:
                recurrence = cal_item.recurrence
                copytags = _copytags(cal_item.mapiobj)

                if delete:
                    recurrence._delete_exception(
                        basedate, self.item, copytags)
                else:
                    if recurrence._is_exception(basedate):
                        recurrence._modify_exception(
                            basedate, self.item, copytags)
                    else:
                        recurrence._create_exception(
                            basedate, self.item, copytags)

                    message = recurrence._exception_message(basedate)
                    message[PidLidBusyStatus] = libfreebusy.fbFree
                    message[PR_MESSAGE_FLAGS] = MSGFLAG_UNSENT | MSGFLAG_READ

                    _utils._save(message._attobj)
            else:
                if delete:
                    self.calendar.delete(cal_item)
                else:
                    cal_item.cancel()
        else:
            if delete:
                self.calendar.delete(cal_item)
            else:
                self.item.mapiobj.CopyTo(
                    [], [], 0, None, IID_IMessage, cal_item.mapiobj, 0)
                cal_item.mapiobj.SetProps(
                    [SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Appointment')])

        if cal_item and not delete:
            _utils._save(cal_item.mapiobj)

    def process_response(self):
        """Process meeting request response."""
        if not self.is_response:
            raise Error('item is not a meeting request response')
        if self._check_processed():
            self.log.warning('response already processed')
            return

        cal_item = self.calendar_item
        basedate = self.basedate

        # modify calendar item or embedded message (in case of exception)
        attach = None
        message = None
        if basedate:
            if cal_item.recurring:
                recurrence = cal_item.recurrence

                recurrence._update_calitem() # TODO via create/modify exception

                if recurrence._is_exception(basedate):
                    message = recurrence._exception_message(basedate)

                    owner_appt_id = self.item.get(PR_OWNER_APPT_ID)
                    if owner_appt_id is not None:
                        message[PR_OWNER_APPT_ID] = owner_appt_id
                    attach = message._attobj
        else:
            message = cal_item
        if not message:
            self.log.debug('no appointment matches response')
            return

        # update recipient track status
        table = message.mapiobj.OpenProperty(
            PR_MESSAGE_RECIPIENTS, IID_IMAPITable, MAPI_UNICODE, 0)
        # TODO things seem to get lost without this
        table.SetColumns(RECIP_PROPS, 0)

        rows = table.QueryRows(2147483647, 0)
        for row in rows:
            if self._compare_ab_entryids(
                PpropFindProp(row, PR_ENTRYID).Value,
                self.item.prop(PR_SENT_REPRESENTING_ENTRYID).value
            ):
                trackstatus = PpropFindProp(row, PR_RECIPIENT_TRACKSTATUS)
                trackstatus_time = PpropFindProp(row, PR_RECIPIENT_TRACKSTATUS_TIME)

                attendee_crit_change = self.item.get_prop(PidLidAttendeeCriticalChange)

                if trackstatus_time and attendee_crit_change and attendee_crit_change.mapiobj.Value <= trackstatus_time.Value:
                    continue

                if trackstatus_time:
                    trackstatus_time.Value = attendee_crit_change.mapiobj.Value
                else:
                    row.append(SPropValue(PR_RECIPIENT_TRACKSTATUS_TIME,
                                          attendee_crit_change.mapiobj.Value))

                if trackstatus:
                    trackstatus.Value = self.track_status
                else:
                    row.append(SPropValue(
                        PR_RECIPIENT_TRACKSTATUS, self.track_status))

                if self.item.get(PidLidAppointmentCounterProposal) is True:
                    start = self.item.prop(PidLidAppointmentProposedStartWhole)
                    end = self.item.prop(PidLidAppointmentProposedEndWhole)
                    row.extend([
                        SPropValue(PR_PROPOSED_NEWTIME, True),
                        SPropValue(PR_PROPOSED_NEWTIME_START, start.mapiobj.Value),
                        SPropValue(PR_PROPOSED_NEWTIME_END, end.mapiobj.Value),
                    ])

        message.mapiobj.ModifyRecipients(MODRECIP_MODIFY, rows)

        # counter proposal
        if self.item.get(PidLidAppointmentCounterProposal):
            message[PidLidAppointmentCounterProposal] = True

        # save all the things
        _utils._save(message.mapiobj)

        if attach:
            _utils._save(attach)
            _utils._save(cal_item.mapiobj)

    def _compare_ab_entryids(self, entryid1, entryid2):  # TODO shorten?
        smtp1 = self._get_smtp_address(entryid1)
        smtp2 = self._get_smtp_address(entryid2)
        return smtp1 == smtp2

    def _get_smtp_address(self, entryid):
        ab = self.item.server.ab
        try:
            abitem = ab.OpenEntry(entryid, None, 0)
        except MAPIErrorUnknownEntryid:
            return ''

        props = abitem.GetProps(
            [PR_ADDRTYPE_W, PR_EMAIL_ADDRESS_W, PR_SMTP_ADDRESS_W], 0)
        if props[0].Value == 'SMTP':
            return props[1].Value
        else:
            return props[2].Value

    def __unicode__(self):
        return 'MeetingRequest()'

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-only
import datetime

# PT_SYSTIME properties store the number of 100-nanosecond units since 1601,1,1 ..
# unixtime is represented by the number of seconds since 1970,1,1

# number of 100-nanosecond units between 1601,1,1 and 1970,1,1
NANOSECS_BETWEEN_EPOCH = 116444736000000000

def _convert(s):
    if isinstance(s, bytes):
        return s.decode('ascii')
    else:
        return s


def datetime_to_filetime(d):
    return FileTime(int(datetime.datetime.timestamp(d)) * 10000000 + NANOSECS_BETWEEN_EPOCH)


# class representing a PT_SYSTIME value. the 'unixtime' property can be used to convert to/from unixtime.

class FileTime(object):
    def __init__(self, filetime):
        self.filetime = filetime

    def datetime(self):
        return datetime.datetime.utcfromtimestamp(self.unixtime)

    def __getattr__(self, attr):
        if attr == 'unixtime':
            return (self.filetime - NANOSECS_BETWEEN_EPOCH) / 10000000.0;
        else:
            raise AttributeError

    def __setstate__(self, d):
        # XXX pickle with python2, unpickle with python3 (encoding='bytes')
        for k, v in d.items():
            setattr(self, _convert(k), v)

    def __setattr__(self, attr, val):
        if attr == 'unixtime':
            self.filetime = val * 10000000 + NANOSECS_BETWEEN_EPOCH
        else:
            object.__setattr__(self, attr, val)

    def __repr__(self):
        try:
            return self.datetime().isoformat()
        except ValueError:
            return '%d' % (self.filetime)

    def __lt__(self, other):
        return self.filetime < other.filetime
    def __le__(self, other):
        return self.filetime <= other.filetime
    def __gt__(self, other):
        return self.filetime > other.filetime
    def __ge__(self, other):
        return self.filetime >= other.filetime
    def __eq__(self, other):
        return isinstance(other, FileTime) and self.filetime == other.filetime
    def __ne__(self, other):
        return not isinstance(other, FileTime) or self.filetime != other.filetime

# convert unixtime to PT_SYSTIME.. (bad name, as it sounds like the result is a unixtime)

def unixtime(secs):
    t = FileTime(0)
    t.unixtime = secs
    return t



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI.Struct import ECQUOTA, MAPIErrorNotFound, MAPIErrorCollision

from .defs import CONTAINER_COMPANY, ACTIVE_USER
from .errors import NotFoundError, DuplicateError

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

class Quota(object):
    """Quota class

    Manage :class:`user <User>` or :class:`company <Company>`
    quota settings.
    """

    def __init__(self, server, userid):
        self.server = server
        self.userid = userid
        # XXX quota for 'default' company?
        self._warning_limit = self._soft_limit = self._hard_limit = 0
        if userid:
            quota = server.sa.GetQuota(userid, False)
            self._warning_limit = quota.llWarnSize
            self._soft_limit = quota.llSoftSize
            self._hard_limit = quota.llHardSize
            # XXX: logical name for variable
            # Use default quota set in /etc/kopano/server.cfg
            self._use_default_quota = bool(quota.bUseDefaultQuota)
            # XXX: is this for multitenancy?
            self._isuser_default_quota = quota.bIsUserDefaultQuota

    @property
    def warning_limit(self):
        """Warning limit."""
        return self._warning_limit

    @warning_limit.setter
    def warning_limit(self, value):
        self.update(warning_limit=value)

    @property
    def soft_limit(self):
        """Soft limit."""
        return self._soft_limit

    @soft_limit.setter
    def soft_limit(self, value):
        self.update(soft_limit=value)

    @property
    def hard_limit(self):
        """Hard limit."""

        return self._hard_limit

    @hard_limit.setter
    def hard_limit(self, value):
        self.update(hard_limit=value)

    #TODO: support defaultQuota and IsuserDefaultQuota
    def update(self, **kwargs):
        """
        Update function for Quota limits.

        :param warning_limit: Warning limit.
        :param soft_limit: Soft limit.
        :param hard_limit: Hard limit.
        """
        self._warning_limit = kwargs.get('warning_limit', self._warning_limit)
        self._soft_limit = kwargs.get('soft_limit', self._soft_limit)
        self._hard_limit = kwargs.get('hard_limit', self._hard_limit)
        self._use_default_quota = kwargs.get('use_default',
            self._use_default_quota)
        # TODO: implement setting defaultQuota, userdefaultQuota
        # (self, bUseDefaultQuota, bIsUserDefaultQuota, llWarnSize, llSoftSize,
        # llHardSize)
        quota = ECQUOTA(self._use_default_quota, False, self._warning_limit,
            self._soft_limit, self._hard_limit)
        self.server.sa.SetQuota(self.userid, quota)

    @property
    def use_default(self):
        """Use default quota."""
        return self._use_default_quota

    @use_default.setter
    def use_default(self, x):
        self.update(use_default=x)

    def recipients(self):
        """Return all :class:`recipients <User>` of quota messages."""
        if self.userid:
            for ecuser in self.server.sa.GetQuotaRecipients(self.userid, 0):
                yield self.server.user(ecuser.Username)

    # XXX remove company flag
    def add_recipient(self, user, company=False):
        """Add :class:`recipient <User>` of quota messages."""
        objclass = CONTAINER_COMPANY if company else ACTIVE_USER
        try:
            self.server.sa.AddQuotaRecipient(self.userid, user._ecuser.UserID,
                objclass)
        except MAPIErrorCollision:
            raise DuplicateError("user '%s' already in %squota recipients" % \
                (user.name, 'company' if company else 'user'))

    # XXX remove company flag
    def remove_recipient(self, user, company=False):
        """Remove :class:`recipient <User>` of quota messages."""
        objclass = CONTAINER_COMPANY if company else ACTIVE_USER
        try:
            self.server.sa.DeleteQuotaRecipient(
                self.userid, user._ecuser.UserID, objclass)
        except MAPIErrorNotFound:
            raise NotFoundError("user '%s' not in %squota recipients" % \
                (user.name, 'company' if company else 'user'))

    def __unicode__(self):
        return 'Quota(warning=%s, soft=%s, hard=%s)' % (
            _utils.bytes_to_human(self.warning_limit),
            _utils.bytes_to_human(self.soft_limit),
            _utils.bytes_to_human(self.hard_limit)
        )

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2017 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI import (
    ROW_REMOVE, FL_PREFIX, RELOP_NE, ROW_ADD, MAPI_BEST_ACCESS,
    MAPI_UNICODE, MAPI_TO,
)
from MAPI.Tags import (
    PR_RULE_CONDITION, PR_RULE_ACTIONS, PR_RULE_PROVIDER_W, ACTTYPE,
    PR_ENTRYID, PR_RULE_ID, PR_RULES_TABLE, IID_IExchangeModifyTable,
    PR_RULE_LEVEL, PR_RULE_NAME_W, PR_RULE_SEQUENCE, PR_RULE_STATE,
    PR_RULE_PROVIDER_DATA, PR_MESSAGE_CLASS_W, PR_DELEGATED_BY_RULE,
    PR_SENSITIVITY, PR_ADDRTYPE_W, PR_EMAIL_ADDRESS_W, PR_DISPLAY_NAME_W,
    PR_SEARCH_KEY, PR_SMTP_ADDRESS_W, PR_OBJECT_TYPE, PR_DISPLAY_TYPE,
    PR_RECIPIENT_TYPE,
)
from MAPI.Defs import (
    PpropFindProp
)
from MAPI.Struct import (
    ROWENTRY, SPropValue, ACTION, actFwdDelegate, ACTIONS, SAndRestriction,
    SOrRestriction, SNotRestriction, SContentRestriction, SPropertyRestriction,
    SExistRestriction,
)
from .compat import (
    bdec as _bdec
)
from .errors import NotFoundError

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

USERPROPS = [
    PR_ENTRYID,
    PR_ADDRTYPE_W,
    PR_EMAIL_ADDRESS_W,
    PR_DISPLAY_NAME_W,
    PR_SEARCH_KEY,
    PR_SMTP_ADDRESS_W,
    PR_OBJECT_TYPE,
    PR_DISPLAY_TYPE,
]

class Delegation(object):
    """Delegation class

    Abstraction for delegate users and settings.
    """

    def __init__(self, store, user):
        self.store = store
        self._user = user

    @property
    def user(self):
        """Delegate user."""
        return self._user

    @property
    def see_private(self):
        """Delegate user can see private :class:`items <Item>`."""
        fbmsg, (entryids, names, flags) = self.store._fbmsg_delgs()
        pos = entryids.Value.index(_bdec(self.user.userid))

        return bool(flags.Value[pos] & 1)

    @see_private.setter
    def see_private(self, b):
        fbmsg, (entryids, names, flags) = self.store._fbmsg_delgs()
        pos = entryids.Value.index(_bdec(self.user.userid))

        if b:
            flags.Value[pos] |= 1
        else:
            flags.Value[pos] &= ~1

        fbmsg.SetProps([flags])
        _utils._save(fbmsg)

    @staticmethod
    def _parse_rule(store):
        userids, deletion = [], False
        for rule in store.inbox.rules():
            if (PR_RULE_PROVIDER_W in rule.mapirow and \
                PR_RULE_ACTIONS in rule.mapirow):
                if rule.mapirow[PR_RULE_PROVIDER_W] == \
                        'Schedule+ EMS Interface':
                    actions = rule.mapirow[PR_RULE_ACTIONS].lpAction
                    if actions and actions[0].acttype == ACTTYPE.OP_DELEGATE:
                        for addrentry in actions[0].actobj.lpadrlist:
                            entryid = PpropFindProp(addrentry, PR_ENTRYID)
                            if entryid:
                                userids.append(entryid.Value)
                    if len(actions) >= 2 and actions[1].acttype == \
                            ACTTYPE.OP_DELETE:
                        deletion = True
        return userids, deletion

    @staticmethod
    def _save_rule(store, userids, deletion):
        # remove existing rule # TODO update
        for rule in store.inbox.rules():
            if (rule.mapirow[PR_RULE_PROVIDER_W] == \
                'Schedule+ EMS Interface' and PR_RULE_ID in rule.mapirow):
                pr_rule_id = rule.mapirow[PR_RULE_ID]

                rulerows = [ROWENTRY(ROW_REMOVE,
                    [SPropValue(PR_RULE_ID, pr_rule_id)])]
                table = store.inbox.mapiobj.OpenProperty(PR_RULES_TABLE,
                    IID_IExchangeModifyTable, 0, 0)
                table.ModifyTable(0, rulerows)

        # create new rule
        row = [
            SPropValue(PR_RULE_LEVEL, 0),
            SPropValue(PR_RULE_NAME_W, 'Delegate Meetingrequest service'),
            SPropValue(PR_RULE_PROVIDER_W, 'Schedule+ EMS Interface'),
            SPropValue(PR_RULE_SEQUENCE, 0),
            SPropValue(PR_RULE_STATE, 1),
            SPropValue(PR_RULE_PROVIDER_DATA, b''),
        ]

        actions = []
        userprops = []
        for userid in userids:
            user = store.server.gab.OpenEntry(userid, None, MAPI_BEST_ACCESS)
            props = user.GetProps(USERPROPS, MAPI_UNICODE)
            # Hardcode recipient type to TO
            props.append(SPropValue(PR_RECIPIENT_TYPE, MAPI_TO))
            userprops.append(props)

        actions.append(ACTION(ACTTYPE.OP_DELEGATE, 0, None, None, 0,
            actFwdDelegate(userprops)))
        if deletion:
            actions.append(ACTION(ACTTYPE.OP_DELETE, 0, None, None, 0, None))
        row.append(SPropValue(PR_RULE_ACTIONS, ACTIONS(1, actions)))

        cond = SAndRestriction([
            SContentRestriction(FL_PREFIX, PR_MESSAGE_CLASS_W,
                SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Schedule.Meeting')),
            SNotRestriction(SExistRestriction(PR_DELEGATED_BY_RULE) ),
            SOrRestriction([
                SNotRestriction(SExistRestriction(PR_SENSITIVITY)),
                SPropertyRestriction(RELOP_NE, PR_SENSITIVITY,
                    SPropValue(PR_SENSITIVITY, 2))
            ])
        ])
        row.append(SPropValue(PR_RULE_CONDITION, cond))
        rulerows = [ROWENTRY(ROW_ADD, row)]
        table = store.inbox.mapiobj.OpenProperty(PR_RULES_TABLE,
            IID_IExchangeModifyTable, 0, 0)
        table.ModifyTable(0, rulerows)

    @property
    def send_copy(self):
        """Delegate user receives copies of meeting requests."""
        userids, deletion = self._parse_rule(self.store)
        return _bdec(self.user.userid) in userids

    @send_copy.setter
    def send_copy(self, value):
        userids, deletion = self._parse_rule(self.store)
        if value:
            userids.append(_bdec(self.user.userid)) # TODO dupe
        else:
            userids = [u for u in userids if u != _bdec(self.user.userid)]
        self._save_rule(self.store, userids, deletion)

    @property
    def flags(self):
        """Delegation flags (see_private, send_copy)."""
        flags = []
        if self.see_private:
            flags.append('see_private')
        if self.send_copy:
            flags.append('send_copy')
        return flags

    @flags.setter
    def flags(self, value):
        self.see_private = ('see_private' in value)
        self.send_copy = ('send_copy' in value)

    @staticmethod
    def _send_only_to_delegates(store):
        """Delete meetingrequests after copying them to delegates."""
        _, deletion = Delegation._parse_rule(store)
        return deletion

    @staticmethod
    def _set_send_only_to_delegates(store, value):
        userids, deletion = Delegation._parse_rule(store)
        Delegation._save_rule(store, userids, value)

    def _delete(self):
        # TODO update delegate rule

        fbmsg, (entryids, names, flags) = self.store._fbmsg_delgs()
        try:
            pos = entryids.Value.index(_bdec(self.user.userid))
        except ValueError:
            raise NotFoundError("no delegation for user '%s'" % self.user.name)

        del entryids.Value[pos]
        del names.Value[pos]
        del flags.Value[pos]

        fbmsg.SetProps([entryids, names, flags])
        _utils._save(fbmsg)

    def __unicode__(self):
        return "Delegation('%s')" % self.user.name

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2018 - 2019 Kopano and its licensors (see LICENSE file)
"""

import io
import logging
import mimetypes


WITH_PIL = False
try:
    from PIL import Image, PngImagePlugin
    WITH_PIL = True
except ImportError: # pragma: no cover
    pass


if WITH_PIL:
    # Get rid of "STREAM" debug messages which might show up for PNG pictures,
    # when the global logger is at DEBUG level.
    logging.getLogger(PngImagePlugin.__name__).setLevel(logging.INFO)

class Picture(object):
    """Picture class

    Typical pictures are :class:`contact <Contact>` and :class:`user <User>`
    profile photos.
    """

    def __init__(self, data, name='', mimetype=''):
        #: Picture name.
        self.name = name
        #: Picture binary data.
        self.data = data
        self._mimetype = mimetype
        self.__img = None

    @property
    def _img(self):
        if self.__img is None and WITH_PIL:
            self.__img = Image.open(io.BytesIO(self.data))
        return self.__img

    @property
    def mimetype(self):
        """Picture MIME type."""
        if not self._mimetype and self._img is not None:
            ext = self._img.format
            if ext:
                self._mimetype = mimetypes.types_map.get('.'+ext.lower(), '')
        return self._mimetype

    @property
    def width(self):
        """Picture pixel width."""
        return self._img.width

    @property
    def height(self):
        """Picture pixel height."""
        return self._img.height

    @property
    def size(self):
        """Picture pixel (width, height) tuple."""
        return (self.width, self.height)

    def scale(self, size):
        """Return new picture instance, scaled to given size.

        :param size: (width, height) tuple
        """
        if not WITH_PIL:
            raise NotImplementedError('PIL is not available')
        img = Image.open(io.BytesIO(self.data))
        img.thumbnail(size)
        b = io.BytesIO()
        img.save(b, format=img.format)
        data = b.getvalue()
        return Picture(data, self.name, self.mimetype)

    def __unicode__(self):
        return 'Picture(%s)' % (self.name or '')

    def __repr__(self):
        return self.__unicode__()



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-only

import sys
import logging

from MAPI import (
    PT_UNICODE
)
from MAPI.Tags import (
    PR_EC_IMAP_EMAIL_SIZE, PR_EC_IMAP_BODYSTRUCTURE_W,
    PR_EC_IMAP_BODY_W, PR_EC_IMAP_EMAIL,
)
from MAPI.Struct import (
    SNotRestriction, SExistRestriction,
)
import inetmapi

from kopano import Restriction, server, parser


def logger(options):
    logging.basicConfig(stream=sys.stdout, level=options.loglevel)
    return logging.getLogger('optimize-imap')


def generate_imap_message(item):
    eml = item.eml(received_date=True)
    envelope, body, bodystructure = inetmapi.createIMAPProperties(eml)

    item.create_prop(PR_EC_IMAP_EMAIL, eml)
    item.create_prop(PR_EC_IMAP_EMAIL_SIZE, len(eml))
    item.create_prop(PR_EC_IMAP_BODYSTRUCTURE_W, bodystructure)
    item.create_prop(PR_EC_IMAP_BODY_W, body)
    item.create_prop('imap:1', envelope, proptype=PT_UNICODE)

    item.mapiobj.SaveChanges(0)

def main():
    options, _ = parser('ksplu').parse_args()
    log = logger(options)
    server = server(options=options, auth_user='SYSTEM', auth_pass='', parse_args=True)
    restriction = Restriction(mapiobj=SNotRestriction(SExistRestriction(PR_EC_IMAP_EMAIL_SIZE)))

    for user in server.users():  # XXX multi-company..
        # Skip users without IMAP enabled
        if 'imap' not in user.features:
            log.info('Skipping user %s, IMAP disabled', user.name)
            continue

        log.debug('Processing user %s', user.name)
        for folder in user.store.folders():
            # Inbox folder's container class is None..
            if folder.container_class != 'IPF.Note' and folder.container_class is not None:
                continue

            log.info('Processing folder %s', folder.name)
            for item in folder.items(restriction=restriction):
                log.debug('Processing item %s', item.subject)
                generate_imap_message(item)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: AGPL-3.0-only
import struct
import MAPI.Struct

EC_ADDRTYPE = 'ZARAFA'
EC_ADDRTYPE_W = u'ZARAFA'

def PROP_TAG(type, id):
    return id << 16 | type

def PROP_TYPE(x):
    return x & 0xffff

def PROP_ID(x):
    return x >> 16

def CHANGE_PROP_TYPE(tag, type):
    return PROP_TAG(type, PROP_ID(tag))

def DEFINE_GUID(l,w1,w2,b1,b2,b3,b4,b5,b6,b7,b8):
    return struct.pack("I2H8B", l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8)

def DEFINE_OLEGUID(l,w1,w2):
    return DEFINE_GUID(l, w1, w2, 0xC0, 0, 0, 0, 0, 0, 0, 0x46)

def PpropFindProp(props, proptag):
    PT_UNSPECIFIED = 0x0000
    if PROP_TYPE(proptag) == PT_UNSPECIFIED:
        propid = PROP_ID(proptag)
        for prop in props:
            if PROP_ID(prop.ulPropTag) == propid:
                return prop
    else:
        for prop in props:
            if prop.ulPropTag == proptag:
                return prop
    return None

def HrGetOneProp(pmp, proptag):
    props = pmp.GetProps([proptag], 0)
    if props[0].ulPropTag == proptag:
        return props[0]
    raise MAPI.Struct.MAPIError.from_hresult(props[0].Value)
